/*****************************************************************************/
/*                                                                           */
/*  Module Name:     checkout.c                                              */
/*                                                                           */
/*  Description:     This is a standard C program which illustrates the      */
/*                   use of the IBM Point Of Sale Subsystem                  */
/*                   Application Programming Interface (API).                */
/*                                                                           */
/*  Notes:           The program requires 2 arguments:                       */
/*                   the name of a resource file and the name of             */
/*                   a transaction log file.                                 */
/*                                                                           */
/*                   The program expects the item file, checkout.dat,        */
/*                   to be located in the current directory.                 */
/*                                                                           */
/*                   Standard error should be routed to a file               */
/*                   if the touch screen interface is used.                  */
/*                                                                           */
/*                                                                           */
/*      73G0950 (C) Copyright IBM Corporation 1993, 2001                     */
/*                                                                           */
/*      DISCLAIMER OF WARRANTIES.  The following [enclosed] code is          */
/*      sample code created by IBM Corporation.  This sample code is not     */
/*      part of any standard or IBM product and is provided to you solely    */
/*      for the purpose of assisting you in the development of your          */
/*      applications.  The code is provided "AS IS", without warranty of     */
/*      any kind.  IBM shall not be liable for any damages arising out of    */
/*      your use of the sample code, even if they have been advised of the   */
/*      possibility of such damages.                                         */
/*                                                                           */
/*                                                                           */
/*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <ctype.h>
#include <pos/pos.h>
#include <pos/device.h>
#include <pos/display.h>
#include <pos/msr.h>
#include <pos/nvram.h>
#include <pos/power.h>
#include <pos/rs232c.h>
#include <pos/scale.h>
#include <pos/scanner.h>
#include <pos/keyboard.h>
#include <pos/drawer.h>
#include <pos/printer.h>
#include <pos/errno.h>
#include "checkout.h"
#if defined(POSTOUCH)
#include <pos/touch.h>
#include "touchman.h"
#endif

#if !defined(FALSE)
#define FALSE 0
#define TRUE  1
#endif

#define MSR_Buf_Size          300
#define SCN_Buf_Size          300

#define MSR_Hdr_Size          sizeof(PosMsrDataHdr)
#define SCN_Hdr_Size          sizeof(PosScannerDataHdr)
#define Max_Print_Width       36
#define Sleep_ms              100L

#define Tax_Percent           6
#define AUTO_SIGNON           1

/*****************************************************************************/
/*                                                                           */
/*  Program States                                                           */
/*                                                                           */
/*****************************************************************************/
#define Signed_Off_State      1
#define Transaction_State     2
#define Keyboard_Item_State   3
#define Scanner_Item_State    4
#define Tender_State          5
#define Keyboard_Tender_State 6
#define MSR_Tender_State      7
#define Cash_Tendered_State   8
#define Till_Opened_State     9
#define Till_Closed_State     10
#define Store_Closing_State   11
#define Printer_Error_State   12
#define Check_Tender_State    13
#define Check_Remove_State    14
#define Print_Features_State  15


/*****************************************************************************/
/*                                                                           */
/*  Key mappings.  These #defines are used in the keyboard mapping           */
/*                 tables below.                                             */
/*                                                                           */
/*****************************************************************************/
#define SIGN_ON               1
#define ENTER                 2
#define CLEAR                 3
#define CHECK                 4
#define CASH                  5
#define CHARGE                6
#define TOTAL                 7
#define END_PRG               8
#define PRINT                 9
#define WEIGH                10
#define NUM0                 '0'
#define NUM1                 '1'
#define NUM2                 '2'
#define NUM3                 '3'
#define NUM4                 '4'
#define NUM5                 '5'
#define NUM6                 '6'
#define NUM7                 '7'
#define NUM8                 '8'
#define NUM9                 '9'

unsigned char *sioKeys;

/*****************************************************************************/
/*                                                                           */
/*  Checkout Keyboards:                                                      */
/*                                                                           */
/*     Checkout Keyboard                                                     */
/*     Combined Keyboard/Display                                             */
/*                                                                           */
/*****************************************************************************/
unsigned char checkoutKeys[] =
 {
/*   x0,   x1,   x2,      x3       x4,     x5,    x6,      x7                */
/*   x8,   x9,   xA,      xB       xC,     xD,    xE,      xF                */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x00-0x07 */
   0x00, 0x00, 0x00,    0x00,    NUM2,   0x00,  NUM8,    NUM5,  /* 0x08-0x0F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x10-0x17 */
   0x00, 0x00, 0x00,    0x00,    NUM3,   0x00,  NUM9,    NUM6,  /* 0x18-0x1F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x20-0x27 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x28-0x2F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x30-0x37 */
   0x00, 0x00, 0x00,    WEIGH,   0x00,   0x00,  0x00,    0x00,  /* 0x38-0x3F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x40-0x47 */
   0x00, 0x00, 0x00,    END_PRG, 0x00,   CLEAR, 0x00,    0x00,  /* 0x48-0x4F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x50-0x57 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x58-0x5F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x60-0x67 */
   0x00, 0x00, 0x00,    PRINT,   CHECK,  TOTAL, CHARGE,  CASH,  /* 0x68-0x6F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x70-0x77 */
   0x00, 0x00, 0x00,    SIGN_ON, NUM1,   NUM0,  NUM7,    NUM4,  /* 0x78-0x7F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x80-0x87 */
   0x00, 0x00, 0x00,    0x00,    0x00,   ENTER, 0x00,    0x00,  /* 0x88-0x8F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x90-0x97 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x98-0x9F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xA0-0xA7 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xA8-0xAF */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xB0-0xB7 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xB8-0xBF */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xC0-0xC7 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xC8-0xCF */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xD0-0xD7 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xD8-0xDF */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xE0-0xE7 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xE8-0xEF */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xF0-0xF7 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xF8-0xFF */
 };

/*****************************************************************************/
/*                                                                           */
/*  Checkout Keyboards:                                                      */
/*                                                                           */
/*     Retail Point of Sale Keyboard                                         */
/*     Retail Point of Sale Keyboard with Card Reader                        */
/*     Retail Point of Sale Keyboard with Card Reader and Display            */
/*                                                                           */
/*****************************************************************************/
unsigned char checkout2Keys[] =
 {
/*   x0,   x1,   x2,      x3       x4,     x5,    x6,      x7                */
/*   x8,   x9,   xA,      xB       xC,     xD,    xE,      xF                */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x00-0x07 */
   0x00, 0x00, 0x00,    0x00,    NUM8,   0x00,  NUM2,    NUM5,  /* 0x08-0x0F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x10-0x17 */
   0x00, 0x00, 0x00,    0x00,    NUM9,   0x00,  NUM3,    NUM6,  /* 0x18-0x1F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x20-0x27 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x28-0x2F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x30-0x37 */
   0x00, 0x00, 0x00,    WEIGH,   0x00,   0x00,  0x00,    0x00,  /* 0x38-0x3F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x40-0x47 */
   0x00, 0x00, 0x00,    END_PRG, 0x00,   CLEAR, 0x00,    0x00,  /* 0x48-0x4F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x50-0x57 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x58-0x5F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x60-0x67 */
   0x00, 0x00, 0x00,    PRINT,   CHECK,  TOTAL, CHARGE,  CASH,  /* 0x68-0x6F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x70-0x77 */
   0x00, 0x00, 0x00,    SIGN_ON, NUM7,   NUM0,  NUM1,    NUM4,  /* 0x78-0x7F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x80-0x87 */
   0x00, 0x00, 0x00,    0x00,    0x00,   ENTER, 0x00,    0x00,  /* 0x88-0x8F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x90-0x97 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x98-0x9F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xA0-0xA7 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xA8-0xAF */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xB0-0xB7 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xB8-0xBF */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xC0-0xC7 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xC8-0xCF */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xD0-0xD7 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xD8-0xDF */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xE0-0xE7 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xE8-0xEF */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xF0-0xF7 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xF8-0xFF */
 };

/*****************************************************************************/
/*                                                                           */
/*  Modifiable Layout Keyboard with Card Reader                              */
/*                                                                           */
/*****************************************************************************/
unsigned char modifiableKeys[] =
 {
/*   x0,   x1,   x2,      x3       x4,     x5,    x6,      x7                */
/*   x8,   x9,   xA,      xB       xC,     xD,    xE,      xF                */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x00-0x07 */
   0x00, CASH, CHECK,   0x00,    CHARGE, TOTAL, 0x00,    0x00,  /* 0x08-0x0F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x10-0x17 */
   0x00, NUM4, NUM7,    SIGN_ON, NUM1,   0x00,  0x00,    0x00,  /* 0x18-0x1F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x20-0x27 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x28-0x2F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x30-0x37 */
   0x00, 0x00, 0x00,    WEIGH,   0x00,   ENTER, 0x00,    0x00,  /* 0x38-0x3F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x40-0x47 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x48-0x4F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x50-0x57 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x58-0x5F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x60-0x67 */
   0x00, 0x00, 0x00,    END_PRG, 0x00,   CLEAR, 0x00,    0x00,  /* 0x68-0x6F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x70-0x77 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x78-0x7F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x80-0x87 */
   0x00, NUM5, NUM8,    0x00,    NUM2,   NUM0,  0x00,    0x00,  /* 0x88-0x8F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x90-0x97 */
   0x00, NUM6, NUM9,    0x00,    NUM3,   0x00,  0x00,    0x00,  /* 0x98-0x9F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xA0-0xA7 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xA8-0xAF */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xB0-0xB7 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xB8-0xBF */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xC0-0xC7 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xC8-0xCF */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xD0-0xD7 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xD8-0xDF */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xE0-0xE7 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xE8-0xEF */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xF0-0xF7 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xF8-0xFF */
 };

/*****************************************************************************/
/*                                                                           */
/*  Alphanumeric Point of Sale Keyboard  (SIO)                               */
/*  Retail Point of Sale Alphanumeric Keyboard with Card Reader (SIO)        */
/*                                                                           */
/*****************************************************************************/
unsigned char anposKeys[] =
 {
/*   x0,   x1,   x2,      x3       x4,     x5,    x6,      x7                */
/*   x8,   x9,   xA,      xB       xC,     xD,    xE,      xF                */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x00-0x07 */
   0x00, 0x00, 0x00,    NUM4,    NUM5,   0x00,  CHARGE,  NUM6,  /* 0x08-0x0F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x10-0x17 */
   0x00, 0x00, 0x00,    0x00,    0x00, END_PRG, 0x00,    0x00,  /* 0x18-0x1F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x20-0x27 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x28-0x2F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x30-0x37 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    CASH,  /* 0x38-0x3F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x40-0x47 */
   0x00, 0x00, ENTER,   0x00,    0x00,   0x00,  WEIGH,   0x00,  /* 0x48-0x4F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x50-0x57 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    NUM3,  /* 0x58-0x5F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x60-0x67 */
   0x00, 0x00, 0x00,    NUM7,    NUM8,   0x00,  CHECK,  NUM9,  /* 0x68-0x6F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x70-0x77 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x78-0x7F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x80-0x87 */
   0x00, 0x00, 0x00,    NUM1,    NUM2, SIGN_ON, 0x00,    0x00,  /* 0x88-0x8F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0x90-0x97 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  CLEAR,   0x00,  /* 0x98-0x9F */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xA0-0xA7 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xA8-0xAF */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xB0-0xB7 */
   0x00, 0x00, 0x00,    NUM0,    NUM0,   0x00,  TOTAL,   0x00,  /* 0xB8-0xBF */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xC0-0xC7 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xC8-0xCF */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xD0-0xD7 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xD8-0xDF */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xE0-0xE7 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xE8-0xEF */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xF0-0xF7 */
   0x00, 0x00, 0x00,    0x00,    0x00,   0x00,  0x00,    0x00,  /* 0xF8-0xFF */
 };

typedef struct
  {
    long cash;
    long credit;
    long tax;
  } nvram_data;

typedef struct
  {
    char *dev_name;
    char *cls_name;
    int  devdes;
    int  dev_type;
    int  dev_subtype;
    int  dev_number;
  } device, *devicelist;

typedef struct
  {
    char *code;
    long rec_num;
  } item_entry;

typedef struct TableEntry
  {
    int Value;
    char *Name;
  } TableEntry ;

#define DSP                   0
#define MSR                   1
#define SCN                   2
#define SCL                   3
#define KBD                   4
#define TIL                   5
#define PRT                   6 
#define RAM                   7 
#if defined(POSTOUCH)
#define TCH                   8 
#endif

device devInfo[] =
  {
    { "display",      PosDisplay,   -1, 0, 0, 0 },
    { "msr",          PosMsr,       -1, 0, 0, 0 },
    { "scanner",      PosScanner,   -1, 0, 0, 0 },
    { "scale",        PosScale,     -1, 0, 0, 0 },
    { "keyboard",     PosKeyboard,  -1, 0, 0, 0 },
    { "cash_drawer",  PosDrawer,    -1, 0, 0, 0 },
    { "printer",      PosPrinter,   -1, 0, 0, 0 },
    { "hard_totals",  PosNvram,     -1, 0, 0, 0 },
#if defined(POSTOUCH)                          
    { "touch",        PosTouch,     -1, 0, 0, 0 },
#endif
  };
int num_devices = (sizeof(devInfo) / sizeof(devInfo[0]));

int  CashKey(transactionRecord *tran);
int  ChargeKey(transactionRecord *tran);
int  CheckKey(transactionRecord *tran);
int  ClearKey(keyboard *kbd);
void CloseAll(devicelist, int);
void DeviceStatus(POSQMSG *);
void DisplayAmount(int, long);
void DisplayFullScreen(int, char *, int);
void DisplayItemRight(int, char *);
void DisplayString(int, char *, int);
void GetPrinterStatus(int, long *);
int  EndProgramKey(void);
int  EnterKey(transactionRecord *tran, keyboard *kbd);
long ItemLookup(char *, char *);
void LockDevice(int);
int  OpenAndAcquire(int, char *, char *, PosArgPtr, int);
void OpenCashDrawer(int);
void ProcessCharacter(POSQMSG *, transactionRecord *, keyboard *);
int  ProcessNumber(transactionRecord *, keyboard *, unsigned char);
#if defined(POSTOUCH)
void ProcessTouchData(POSQMSG *, transactionRecord *, keyboard *);
#endif
int  ReadMSR(int devdes, char *, int *);
int  ReadMicr(char *data, long *length);
void ReadNVRAM(long *, long *, long *);
int  ReadScanner(int, char *, int *);
void ReleaseAll(devicelist, int);
int  SignOn(transactionRecord *tran, keyboard *kbd);
void SoundTone(void);
int  TotalKey(transactionRecord *tran);
void UnlockDevice(int);
void UpdateNVRAM(int, long, long, long);
int  WriteFeatures(void);
void WritePrinterClose(char *, long);
void WritePrinterDate(int);
void WritePrinterItem(char *, long);
void WriteToCRstation(char *);
void WriteToCRandSJstations(char *);
void WriteToDIstation(char *buffer, size_t length);
void WriteToSJstation(char *);
int  WriteTransactionLog(transactionRecord *);
void ZeroNVRAM(void);

char *ltoa(long l, char *a, size_t a_size);


POSQMSG qmsg;
FILE    *item_rec_file;
int     pos_state         = Signed_Off_State;
int     sav_state         = Signed_Off_State;
int     MicrInfoRequested = FALSE;
int     MicrInstalled     = FALSE;
int     welcome_displayed = FALSE;
int     exit_switch       = FALSE;
int     cash_tend_switch  = FALSE;
int     sjPrintItems      = FALSE;
long    total_amount      = 0;
char    chase_print[]     = { 0x1B, 0x4F, 0x1B, 0x50, 0x1B, 0x47, 0 };
char    achGreeting[41]   = " IBM Point of Sale       Subsystem      ";

/*******************************************************************/
/*    Macros                                                       */
/*******************************************************************/
#define MakeEntry(x) { x , # x }
#define MakeExtEntry(x, y) { x ## y , # y }
#define MakeSubtypeEntry( device, type, subtype )                         \
  { ( PosDEVICE_ ## device << 8 ) | Pos ## type ## _SUBTYPE_ ## subtype , \
    # subtype }


/*******************************************************************/
/* Tables used by GetTableEntry and FindMatch                      */
/*******************************************************************/

TableEntry ErrnoTable[] =
  {
    MakeEntry(POSERR_DSP_INVALID_CURSOR),
    MakeEntry(POSERR_DSP_INVALID_MODE),
    MakeEntry(POSERR_DSP_INVALID_SIZE),
    MakeEntry(POSERR_DSP_UNSUPPORTED_BITMAP),
    MakeEntry(POSERR_DSP_BAD_BITMAP),
    MakeEntry(POSERR_DSP_INVALID_CODE_PAGE),

    MakeEntry(POSERR_KBD_INVALID_FREQUENCY),
    MakeEntry(POSERR_KBD_INVALID_DURATION),
    MakeEntry(POSERR_KBD_INVALID_VOLUME),
    MakeEntry(POSERR_KBD_FILE_NOT_FOUND),
    MakeEntry(POSERR_KBD_INVALID_DOUBLE_KEY),
    MakeEntry(POSERR_KBD_INVALID_FAT_FINGER_TIMEOUT),
    MakeEntry(POSERR_KBD_INVALID_KEYBOARD_CLICK),
    MakeEntry(POSERR_KBD_INVALID_NUMPAD_STYLE),
    MakeEntry(POSERR_KBD_INVALID_NUMPAD_ZERO),
    MakeEntry(POSERR_KBD_INVALID_TYPEMATIC_DELAY),
    MakeEntry(POSERR_KBD_INVALID_TYPEMATIC_FREQ),
    MakeEntry(POSERR_KBD_INVALID_NUMPAD_LOCATION),
    MakeEntry(POSERR_KBD_INVALID_TRAP_KEYS),

    MakeEntry(POSERR_NVRAM_NOT_ENOUGH_ROOM),
    MakeEntry(POSERR_NVRAM_INVALID_CURSOR),
    MakeEntry(POSERR_NVRAM_EOF),
    MakeEntry(POSERR_NVRAM_INVALID_MODE),
    MakeEntry(POSERR_NVRAM_INVALID_LENGTH_RECORD),
    MakeEntry(POSERR_NVRAM_INVALID_DATA_CRC),
    MakeEntry(POSERR_NVRAM_BATTERY_BAD),

    MakeEntry(POSERR_POWER_INVALID_DAY),
    MakeEntry(POSERR_POWER_INVALID_HOUR),
    MakeEntry(POSERR_POWER_INVALID_MINUTES),

    MakeEntry(POSERR_PRN_INVALID_DI_WIDTH),
    MakeEntry(POSERR_PRN_INVALID_INTERLEAVED_VALUE),
    MakeEntry(POSERR_PRN_INVALID_HEAD_PARKED_POSITION),
    MakeEntry(POSERR_PRN_INVALID_STATION),
    MakeEntry(POSERR_PRN_INVALID_MODE),
    MakeEntry(POSERR_PRN_INVALID_CR_LF_DISTANCE),
    MakeEntry(POSERR_PRN_INVALID_DI_LF_DISTANCE),
    MakeEntry(POSERR_PRN_INVALID_SJ_LF_DISTANCE),
    MakeEntry(POSERR_PRN_INVALID_FEED_DIRECTION),
    MakeEntry(POSERR_PRN_INVALID_FISCAL_NOTIFY),
    MakeEntry(POSERR_PRN_INVALID_DI_ORIENTATION),
    MakeEntry(POSERR_PRN_INVALID_LEFT_MARGIN_CR),
    MakeEntry(POSERR_PRN_INVALID_PRINT_ALIGNMENT),
    MakeEntry(POSERR_PRN_INCORRECT_DATA_FORMAT),
    MakeEntry(POSERR_PRN_LOGO_EXISTS),
    MakeEntry(POSERR_PRN_UDC_CHARACTER_EXISTS),
    MakeEntry(POSERR_PRN_INCORRECT_RAW_DATA),
    MakeEntry(POSERR_PRN_INVALID_QUALITY_MODE),
    MakeEntry(POSERR_PRN_INVALID_UPSIDE_DOWN_MODE),
    MakeEntry(POSERR_PRN_INVALID_TABSTOPS_FORMAT),
    MakeEntry(POSERR_PRN_INVALID_COLOR_MODE),
    MakeEntry(POSERR_PRN_INVALID_TONE_VOLUME),
    MakeEntry(POSERR_PRN_INVALID_TONE_DURATION),
    MakeEntry(POSERR_PRN_INVALID_TONE_NOTE),
    MakeEntry(POSERR_PRN_INVALID_TONE_OCTAVE),
    MakeEntry(POSERR_PRN_INVALID_TONE_FREQUENCY),
    MakeEntry(POSERR_PRN_INVALID_CODE_PAGE),

    MakeEntry(POSERR_RS232_INVALID_BAUD_RATE),
    MakeEntry(POSERR_RS232_INVALID_STOP_BITS),
    MakeEntry(POSERR_RS232_INVALID_PARITY),
    MakeEntry(POSERR_RS232_INVALID_DATA_BITS),
    MakeEntry(POSERR_RS232_INVALID_TIMEOUT_CHAR),
    MakeEntry(POSERR_RS232_PREV_NOT_COMPLETE),

    MakeEntry(POSERR_SCALE_INVALID_OPERATIONS_MODE),
    MakeEntry(POSERR_SCALE_INVALID_REMOTE_DISPLAY_STATE),
    MakeEntry(POSERR_SCALE_INVALID_VIBRATION_FILTER),
    MakeEntry(POSERR_SCALE_INVALID_WEIGHT_MODE),
    MakeEntry(POSERR_SCALE_INVALID_ZERO_INDICATOR_STATE),
    MakeEntry(POSERR_SCALE_INVALID_ZERO_RETURN_STATE),
    MakeEntry(POSERR_SCALE_ZERO_SCALE_FAILED),
    MakeEntry(POSERR_SCALE_INVALID_CLEAR_SCREEN_REQUEST),
    MakeEntry(POSERR_SCALE_CONFIGURATION_ERROR),
    MakeEntry(POSERR_SCALE_MSG_LENGTH_ERROR),
    MakeEntry(POSERR_SCALE_INVALID_NUM_WEIGHT_DIGITS),

    MakeEntry(POSERR_SCAN_INVALID_BAR_CODES_1),
    MakeEntry(POSERR_SCAN_INVALID_BAR_CODES_2),
    MakeEntry(POSERR_SCAN_INVALID_BEEP_FREQ),
    MakeEntry(POSERR_SCAN_INVALID_BEEP_LENGTH),
    MakeEntry(POSERR_SCAN_INVALID_BEEP_STATE),
    MakeEntry(POSERR_SCAN_INVALID_BEEP_VOLUME),
    MakeEntry(POSERR_SCAN_INVALID_BLINK_LENGTH),
    MakeEntry(POSERR_SCAN_INVALID_BLOCK_READ_MODE),
    MakeEntry(POSERR_SCAN_INVALID_BLOCK_1_TYPE),
    MakeEntry(POSERR_SCAN_INVALID_BLOCK_2_TYPE),
    MakeEntry(POSERR_SCAN_INVALID_BLOCK_3_TYPE),
    MakeEntry(POSERR_SCAN_INVALID_CHECK_MODULO),
    MakeEntry(POSERR_SCAN_INVALID_D_READ_TIMEOUT),
    MakeEntry(POSERR_SCAN_INVALID_D_TOUCH_MODE),
    MakeEntry(POSERR_SCAN_INVALID_ITF_LENGTH_1),
    MakeEntry(POSERR_SCAN_INVALID_ITF_LENGTH_2),
    MakeEntry(POSERR_SCAN_INVALID_LASER_TIMEOUT),
    MakeEntry(POSERR_SCAN_INVALID_MOTOR_TIMEOUT),
    MakeEntry(POSERR_SCAN_INVALID_LASER_SWITCH_STATE),
    MakeEntry(POSERR_SCAN_INVALID_SCANS_PER_READ),
    MakeEntry(POSERR_SCAN_LABEL_TOO_SHORT),
    MakeEntry(POSERR_SCAN_LABEL_TOO_LONG),
    MakeEntry(POSERR_SCAN_MSG_LENGTH_ERROR),
    MakeEntry(POSERR_SCAN_INVALID_BVOL_SWITCH_STATE),
    MakeEntry(POSERR_SCAN_INVALID_DECODE_ALGORITHM),
    MakeEntry(POSERR_SCAN_INVALID_EAN13_SCANS_PER_READ),
    MakeEntry(POSERR_SCAN_INVALID_EAN8_SCANS_PER_READ),
    MakeEntry(POSERR_SCAN_INVALID_STORE_SCANS_PER_READ),
    MakeEntry(POSERR_SCAN_INVALID_UPCA_SCANS_PER_READ),
    MakeEntry(POSERR_SCAN_INVALID_UPCD_SCANS_PER_READ),
    MakeEntry(POSERR_SCAN_INVALID_UPCE_SCANS_PER_READ),
    MakeEntry(POSERR_SCAN_INVALID_UPC_EXPANSION),
    MakeEntry(POSERR_SCAN_INVALID_VERIFY_PRICE_CHK),
    MakeEntry(POSERR_SCAN_INVALID_QUEUE_ALL_INDICATOR),
    MakeEntry(POSERR_SCAN_CONFIGURATION_ERROR),
    MakeEntry(POSERR_SCAN_2_LABEL_FLAG_CONFIG_ERROR),
    MakeEntry(POSERR_SCAN_INVALID_2_LABEL_DECODE_STATE),
    MakeEntry(POSERR_SCAN_INVALID_FLAG_PAIR_COMBINATION),
    MakeEntry(POSERR_SCAN_INVALID_2_LABEL_FLAG_PAIR_1),
    MakeEntry(POSERR_SCAN_INVALID_2_LABEL_FLAG_PAIR_2),
    MakeEntry(POSERR_SCAN_INVALID_2_LABEL_FLAG_PAIR_3),
    MakeEntry(POSERR_SCAN_INVALID_2_LABEL_FLAG_PAIR_4),
    MakeEntry(POSERR_SCAN_INVALID_CODE39_SCANS_PER_READ),
    MakeEntry(POSERR_SCAN_INVALID_INT2OF5_SCANS_PER_READ),
    MakeEntry(POSERR_SCAN_INVALID_CODE128_SCANS_PER_READ),
    MakeEntry(POSERR_SCAN_INVALID_BAR_CODES_3),
    MakeEntry(POSERR_SCAN_INVALID_BAR_CODES_4),
    MakeEntry(POSERR_SCAN_FLASH_UPDATE_REQD),
    MakeEntry(POSERR_SCAN_INVALID_ITF_LENGTH_TYPE),
    MakeEntry(POSERR_SCAN_INVALID_SUPPLEMENTALS),
    MakeEntry(POSERR_SCAN_INVALID_BARCODE_PROG_STATE),
    MakeEntry(POSERR_SCAN_INVALID_XMIT_CHECK_DIGIT),

    MakeEntry(POSERR_SYS_OS_ERROR),
    MakeEntry(POSERR_SYS_NOT_INITIALIZED),
    MakeEntry(POSERR_SYS_INVALID_DESCRIPTOR),
    MakeEntry(POSERR_SYS_ALREADY_INITIALIZED),
    MakeEntry(POSERR_SYS_MEMORY_ALLOCATION),
    MakeEntry(POSERR_SYS_HW_ERROR),
    MakeEntry(POSERR_SYS_INVALID_DEVICE),
    MakeEntry(POSERR_SYS_INVALID_QUEUE),
    MakeEntry(POSERR_SYS_TOO_MANY_DEVICES),
    MakeEntry(POSERR_SYS_OUT_OF_BOUNDS),
    MakeEntry(POSERR_SYS_FUNCTION_NOT_SUPPORTED),
    MakeEntry(POSERR_SYS_BUFFER_TOO_SMALL),
    MakeEntry(POSERR_SYS_ACQUIRED_BY_OTHER),
    MakeEntry(POSERR_SYS_ALREADY_ACQUIRED),
    MakeEntry(POSERR_SYS_NOT_ACQUIRED),
    MakeEntry(POSERR_SYS_INVALID_REQUEST),
    MakeEntry(POSERR_SYS_DEVICE_OFFLINE),
    MakeEntry(POSERR_SYS_INVALID_LENGTH),
    MakeEntry(POSERR_SYS_INVALID_CLASS_DEVICE_COMBO),
    MakeEntry(POSERR_SYS_DATA_DISCARDED),
    MakeEntry(POSERR_SYS_INTERNAL_ERROR),
    MakeEntry(POSERR_SYS_CMD_REJECT),
    MakeEntry(POSERR_SYS_WRAP_FAILED),
    MakeEntry(POSERR_SYS_DIAG_FAILED),
    MakeEntry(POSERR_SYS_INVALID_NARGS),
    MakeEntry(POSERR_SYS_INVALID_SLOT),
    MakeEntry(POSERR_SYS_UNSUPPORTED_ADAPTER),
    MakeEntry(POSERR_SYS_INVALID_PORT),
    MakeEntry(POSERR_SYS_TIMEOUT),
    MakeEntry(POSERR_SYS_INVALID_NAME),
    MakeEntry(POSERR_SYS_INVALID_CLASS),
    MakeEntry(POSERR_SYS_INTERRUPTED),
    MakeEntry(POSERR_SYS_FILE_ERROR),
    MakeEntry(POSERR_SYS_INVALID_ADDRESS),
    MakeEntry(POSERR_SYS_LOCKED_NO_DATA_READ),
    MakeEntry(POSERR_SYS_INVALID_FILE),
    MakeEntry(POSERR_SYS_SERVICE_NOT_AVAILABLE),

    MakeEntry(POSERR_TILL_INVALID_PULSE_WIDTH),

    MakeEntry(POSERR_TOUCH_INVALID_BACKLIGHT_ON),
    MakeEntry(POSERR_TOUCH_INVALID_CLICK_VOLUME),
    MakeEntry(POSERR_TOUCH_INVALID_CONTRAST),
    MakeEntry(POSERR_TOUCH_INVALID_ENTRY_CLICK),
    MakeEntry(POSERR_TOUCH_INVALID_EXIT_CLICK),
    MakeEntry(POSERR_TOUCH_INVALID_MODE),
    MakeEntry(POSERR_TOUCH_INVALID_SCREEN_SAVER_TIME),
    MakeEntry(POSERR_TOUCH_INVALID_TONE_DURATION),
    MakeEntry(POSERR_TOUCH_INVALID_TONE_FREQUENCY),
    MakeEntry(POSERR_TOUCH_INVALID_TONE_VOLUME),
    MakeEntry(POSERR_TOUCH_INVALID_BRIGHTNESS),
    { 0, 0 }
  };


TableEntry QMsgTable[] =
  {
    MakeEntry(POSM_KBD_STATUS_CHANGE),
    MakeEntry(POSM_KBD_WM_CHAR),
    MakeEntry(POSM_MSR_DATA_AVAIL),
    MakeEntry(POSM_PRN_CHASE_COMPLETE),
    MakeEntry(POSM_PRN_PRINTER_ERROR),
    MakeEntry(POSM_PRN_STATUS_CHANGE),
    MakeEntry(POSM_RS232_BREAK_DETECTED),
    MakeEntry(POSM_RS232_DATA_AVAIL),
    MakeEntry(POSM_RS232_XMIT_ABORT),
    MakeEntry(POSM_RS232_XMIT_COMPLETE),
    MakeEntry(POSM_SCAN_DATA_AVAIL),
    MakeEntry(POSM_SYS_DEVICE_OFFLINE),
    MakeEntry(POSM_SYS_DEVICE_ONLINE),
    MakeEntry(POSM_TILL_CLOSED),
    MakeEntry(POSM_TILL_OPENED),
    MakeEntry(POSM_PRN_DATA_AVAIL),
    MakeEntry(POSM_PRN_FISCAL_ERROR),
    MakeEntry(POSM_PRN_FISCAL_STATUS),
    MakeEntry(POSM_SYS_DEVICE_RELEASED),
    MakeEntry(POSM_PRN_RAS_STATUS),
    MakeEntry(POSM_PRN_RAS_ERROR),
    MakeEntry(POSM_TOUCH_DATA),
    MakeEntry(POSM_PRN_RAW_STATUS),
    MakeEntry(POSM_MSR_WRITE_SUCCESS),
    MakeEntry(POSM_MSR_WRITE_FAILED),
    MakeEntry(POSM_DSP_RAS_STATUS),
    { 0, 0 }
  };


TableEntry IOCtlTable[] =
  {
    MakeEntry(POS_ALARM_SILENCE_ALARM),
    MakeEntry(POS_ALARM_SOUND_ALARM),
    MakeEntry(POS_DSP_CLEAR_SCREEN),
    MakeEntry(POS_DSP_DEFINE_CHARACTERS),
    MakeEntry(POS_KBD_DISABLE_HOT_KEYS),
    MakeEntry(POS_KBD_DISABLE_NUM_LOCK),
    MakeEntry(POS_KBD_DISABLE_POS_FUNCTIONS),
    MakeEntry(POS_KBD_DISABLE_SCROLL_LOCK),
    MakeEntry(POS_KBD_ENABLE_HOT_KEYS),
    MakeEntry(POS_KBD_ENABLE_NUM_LOCK),
    MakeEntry(POS_KBD_ENABLE_POS_FUNCTIONS),
    MakeEntry(POS_KBD_ENABLE_SCROLL_LOCK),
    MakeEntry(POS_KBD_SET_NUM_LOCK_OFF),
    MakeEntry(POS_KBD_SET_NUM_LOCK_ON),
    MakeEntry(POS_KBD_SET_SCROLL_LOCK_OFF),
    MakeEntry(POS_KBD_SET_SCROLL_LOCK_ON),
    MakeEntry(POS_KBD_SET_TYPEMATIC_OFF),
    MakeEntry(POS_KBD_SET_TYPEMATIC_ON),
    MakeEntry(POS_KBD_SILENCE_TONE),
    MakeEntry(POS_KBD_SOUND_TONE),
    MakeEntry(POS_POWER_OFF),
    MakeEntry(POS_POWER_ON),
    MakeEntry(POS_POWER_SET_ALARM),
    MakeEntry(POS_PRN_DISABLE_DI_PRINTING),
    MakeEntry(POS_PRN_DISABLE_FISCAL_PRINTING),
    MakeEntry(POS_PRN_DISCARD_DATA),
    MakeEntry(POS_PRN_ENABLE_DI_PRINTING),
    MakeEntry(POS_PRN_ENABLE_FISCAL_PRINTING),
    MakeEntry(POS_PRN_RESET_PRINTER),
    MakeEntry(POS_PRN_RESUME_PRINTING),
    MakeEntry(POS_PRN_RETRY_PRINTING),
    MakeEntry(POS_PRN_DEFINE_CHARACTERS),
    MakeEntry(POS_PRN_HOLD_PRINTING),
    MakeEntry(POS_PRN_RELEASE_PRINTING),
    MakeEntry(POS_PRN_DISABLE_EXCLUSIVE_DI_PRINTING),
    MakeEntry(POS_PRN_ENABLE_EXCLUSIVE_DI_PRINTING),
    MakeEntry(POS_PRN_SOUND_TONE),
    MakeEntry(POS_PRN_SILENCE_TONE),
    MakeEntry(POS_RS232_SEND_BREAK),
    MakeEntry(POS_SCALE_CLEAR_SCREEN),
    MakeEntry(POS_SCALE_ZERO_SCALE),
    MakeEntry(POS_SCAN_DISCARD_DATA),
    MakeEntry(POS_SYS_ACQUIRE_DEVICE),
    MakeEntry(POS_SYS_GET_VALUES),
    MakeEntry(POS_SYS_LOCK_DEVICE),
    MakeEntry(POS_SYS_RELEASE_DEVICE),
    MakeEntry(POS_SYS_SET_VALUES),
    MakeEntry(POS_SYS_UNLOCK_DEVICE),
    MakeEntry(POS_TILL_OPEN_TILL),
    MakeEntry(POS_TOUCH_SILENCE_TONE),
    MakeEntry(POS_TOUCH_SOUND_TONE),
    { 0, 0 }
  };

TableEntry SlotTable[] =
  {
    MakeExtEntry(Pos, SLOT_0),
    MakeExtEntry(Pos, SLOT_1),
    MakeExtEntry(Pos, SLOT_2),
    MakeExtEntry(Pos, SLOT_3),
    MakeExtEntry(Pos, SLOT_4),
    MakeExtEntry(Pos, SLOT_5),
    MakeExtEntry(Pos, SLOT_6),
    MakeExtEntry(Pos, SLOT_7),
    MakeExtEntry(Pos, SLOT_8),
    { 0, 0 }
  };

TableEntry PortTable[] =
  {
    MakeExtEntry(Pos, PORT_0),
    MakeExtEntry(Pos, PORT_1),
    MakeExtEntry(Pos, PORT_2),
    { 0, 0 }
  };

TableEntry DevTable[] =
  {
    MakeExtEntry(PosDEVICE_, ALARM                      ),
    MakeExtEntry(PosDEVICE_, ALPHANUMERIC_DISPLAY_A     ),
    MakeExtEntry(PosDEVICE_, ALPHANUMERIC_DISPLAY_B     ),
    MakeExtEntry(PosDEVICE_, ANOP_DISPLAY_A             ),
    MakeExtEntry(PosDEVICE_, ANOP_DISPLAY_B             ),
    MakeExtEntry(PosDEVICE_, APA_DISPLAY_A              ),
    MakeExtEntry(PosDEVICE_, APA_DISPLAY_B              ),
    MakeExtEntry(PosDEVICE_, APA_DISPLAY_C              ),
    MakeExtEntry(PosDEVICE_, APA_DISPLAY_D              ),
    MakeExtEntry(PosDEVICE_, ANPOS_KEYBOARD_A           ),
    MakeExtEntry(PosDEVICE_, ANPOS_KEYBOARD_B           ),
    MakeExtEntry(PosDEVICE_, CASH_DRAWER_A              ),
    MakeExtEntry(PosDEVICE_, CASH_DRAWER_B              ),
    MakeExtEntry(PosDEVICE_, CHECKOUT_KEYBOARD_A        ),
    MakeExtEntry(PosDEVICE_, CHECKOUT_KEYBOARD_B        ),
    MakeExtEntry(PosDEVICE_, MATRIX_KEYBOARD_A          ),
    MakeExtEntry(PosDEVICE_, MATRIX_KEYBOARD_B          ),
    MakeExtEntry(PosDEVICE_, MSR_1_TRACK_A              ),
    MakeExtEntry(PosDEVICE_, MSR_1_TRACK_B              ),
    MakeExtEntry(PosDEVICE_, MSR_2_TRACK_A              ),
    MakeExtEntry(PosDEVICE_, MSR_2_TRACK_B              ),
    MakeExtEntry(PosDEVICE_, MSR_3_TRACK_A              ),
    MakeExtEntry(PosDEVICE_, MSR_3_TRACK_B              ),
    MakeExtEntry(PosDEVICE_, NVRAM_A                    ),
    MakeExtEntry(PosDEVICE_, NVRAM_B                    ),
    MakeExtEntry(PosDEVICE_, NVRAM_LOCAL                ),
    MakeExtEntry(PosDEVICE_, OPERATOR_DISPLAY_A         ),
    MakeExtEntry(PosDEVICE_, OPERATOR_DISPLAY_B         ),
    MakeExtEntry(PosDEVICE_, PLU_KEYBOARD               ),
    MakeExtEntry(PosDEVICE_, POS_KEYBOARD_A             ),
    MakeExtEntry(PosDEVICE_, POS_KEYBOARD_B             ),
    MakeExtEntry(PosDEVICE_, POWER_B                    ),
    MakeExtEntry(PosDEVICE_, POWER_LOCAL                ),
    MakeExtEntry(PosDEVICE_, PRINTER_2                  ),
    MakeExtEntry(PosDEVICE_, PRINTER_3                  ),
    MakeExtEntry(PosDEVICE_, PRINTER_FISCAL_3           ),
    MakeExtEntry(PosDEVICE_, PRINTER_4                  ),
    MakeExtEntry(PosDEVICE_, PRINTER_4610               ),
    MakeExtEntry(PosDEVICE_, PRINTER_4689               ),
    MakeExtEntry(PosDEVICE_, PRINTER_4689_THERMAL       ),
    MakeExtEntry(PosDEVICE_, RS232_23A                  ),
    MakeExtEntry(PosDEVICE_, RS232_23B                  ),
    MakeExtEntry(PosDEVICE_, RS232_25A                  ),
    MakeExtEntry(PosDEVICE_, RS232_25B                  ),
    MakeExtEntry(PosDEVICE_, RS232_A                    ),
    MakeExtEntry(PosDEVICE_, RS232_B                    ),
    MakeExtEntry(PosDEVICE_, SCALE_A                    ),
    MakeExtEntry(PosDEVICE_, SCALE                      ),
    MakeExtEntry(PosDEVICE_, SCANNER_A                  ),
    MakeExtEntry(PosDEVICE_, SCANNER_B                  ),
    MakeExtEntry(PosDEVICE_, SHOPPER_DISPLAY_A          ),
    MakeExtEntry(PosDEVICE_, SHOPPER_DISPLAY_B          ),
    MakeExtEntry(PosDEVICE_, SYSTEM_ATTACHED_KBD        ),
    MakeExtEntry(PosDEVICE_, SYSTEM_ATTACHED_MSR        ),
    MakeExtEntry(PosDEVICE_, TEXT_KEYBOARD_A            ),
    MakeExtEntry(PosDEVICE_, TEXT_KEYBOARD_B            ),
    MakeExtEntry(PosDEVICE_, TOUCH_SCREEN               ),
    MakeExtEntry(PosDEVICE_, TOUCH_SCREEN_B             ),
    { 0, 0 }
  };

TableEntry TypeTable[] =
  {
    MakeExtEntry(PosTYPE_, UNKNOWN      ),
    MakeExtEntry(PosTYPE_, ALARM        ),
    MakeExtEntry(PosTYPE_, CASH_DRAWER  ),
    MakeExtEntry(PosTYPE_, DISPLAY      ),
    MakeExtEntry(PosTYPE_, KEYBOARD     ),
    MakeExtEntry(PosTYPE_, MSR          ),
    MakeExtEntry(PosTYPE_, NVRAM        ),
    MakeExtEntry(PosTYPE_, POWER        ),
    MakeExtEntry(PosTYPE_, PRINTER      ),
    MakeExtEntry(PosTYPE_, RS232        ),
    MakeExtEntry(PosTYPE_, SCALE        ),
    MakeExtEntry(PosTYPE_, SCANNER      ),
    MakeExtEntry(PosTYPE_, TOUCH_SCREEN ),
    { 0, 0 }
  };


TableEntry SubtypeTable[] =
  {
    MakeSubtypeEntry(ANOP_DISPLAY_A, DSP  , 2X20_LCD          ),  // 0x24
    MakeSubtypeEntry(ANOP_DISPLAY_A, DSP  , 2X20_LCD_I        ),
    MakeSubtypeEntry(ANOP_DISPLAY_A, DSP  , 2X20_VFD          ),
    MakeSubtypeEntry(ANOP_DISPLAY_A, DSP  , 2_SIDE_VFD        ),
    MakeSubtypeEntry(ANOP_DISPLAY_A, DSP  , 2X20_LCVFD        ),
    MakeSubtypeEntry(ANOP_DISPLAY_A, DSP  , 2_SIDE_LCVFD      ),

    MakeSubtypeEntry(ANOP_DISPLAY_B, DSP  , 2X20_LCD          ),  // 0x25
    MakeSubtypeEntry(ANOP_DISPLAY_B, DSP  , 2X20_LCD_I        ),
    MakeSubtypeEntry(ANOP_DISPLAY_B, DSP  , 2X20_VFD          ),
    MakeSubtypeEntry(ANOP_DISPLAY_B, DSP  , 2_SIDE_VFD        ),
    MakeSubtypeEntry(ANOP_DISPLAY_B, DSP  , 2X20_LCVFD        ),
    MakeSubtypeEntry(ANOP_DISPLAY_B, DSP  , 2_SIDE_LCVFD      ),

    MakeSubtypeEntry(POS_KEYBOARD_A, KBD  , CHECKOUT_2        ),  // 0x1C
    MakeSubtypeEntry(POS_KEYBOARD_A, KBD  , MODIFIABLE_LAYOUT ),
    MakeSubtypeEntry(POS_KEYBOARD_A, KBD  , ANPOS_2           ),
    MakeSubtypeEntry(POS_KEYBOARD_A, KBD  , PC_POS            ),
    MakeSubtypeEntry(POS_KEYBOARD_A, KBD  , KEYBOARD_V        ),

    MakeSubtypeEntry(POS_KEYBOARD_A, KBD  , KEYBOARD_E        ),         // @P1A

    MakeSubtypeEntry(POS_KEYBOARD_A, KBD  , KEYBOARD_VI       ),
    MakeSubtypeEntry(POS_KEYBOARD_A, KBD  , 4685_K01          ),
    MakeSubtypeEntry(POS_KEYBOARD_A, KBD  , 4820_KEYPAD_NO_KEYS),
    MakeSubtypeEntry(POS_KEYBOARD_A, KBD  , 4820_KEYPAD       ),

    MakeSubtypeEntry(POS_KEYBOARD_B, KBD  , CHECKOUT_2        ),  // 0x1D
    MakeSubtypeEntry(POS_KEYBOARD_B, KBD  , MODIFIABLE_LAYOUT ),
    MakeSubtypeEntry(POS_KEYBOARD_B, KBD  , ANPOS_2           ),
    MakeSubtypeEntry(POS_KEYBOARD_B, KBD  , PC_POS            ),
    MakeSubtypeEntry(POS_KEYBOARD_B, KBD  , KEYBOARD_V        ),
    MakeSubtypeEntry(POS_KEYBOARD_B, KBD  , KEYBOARD_VI       ),
    MakeSubtypeEntry(POS_KEYBOARD_B, KBD  , 4685_K01          ),
    MakeSubtypeEntry(POS_KEYBOARD_B, KBD  , 4820_KEYPAD_NO_KEYS),
    MakeSubtypeEntry(POS_KEYBOARD_B, KBD  , 4820_KEYPAD       ),

    MakeSubtypeEntry(APA_DISPLAY_A,  DSP  , APA_JAPAN         ),  // 0x2A
    MakeSubtypeEntry(APA_DISPLAY_A,  DSP  , APA_KOREA         ),
    MakeSubtypeEntry(APA_DISPLAY_A,  DSP  , APA_TRAD_CHINESE  ),
    MakeSubtypeEntry(APA_DISPLAY_A,  DSP  , APA_SIMP_CHINESE  ),

    MakeSubtypeEntry(APA_DISPLAY_B,  DSP  , APA_JAPAN         ),  // 0x2B
    MakeSubtypeEntry(APA_DISPLAY_B,  DSP  , APA_KOREA         ),
    MakeSubtypeEntry(APA_DISPLAY_B,  DSP  , APA_TRAD_CHINESE  ),
    MakeSubtypeEntry(APA_DISPLAY_B,  DSP  , APA_SIMP_CHINESE  ),

    MakeSubtypeEntry(APA_DISPLAY_C,  DSP  , APA_JAPAN         ),  // 0x2C
    MakeSubtypeEntry(APA_DISPLAY_C,  DSP  , APA_KOREA         ),
    MakeSubtypeEntry(APA_DISPLAY_C,  DSP  , APA_TRAD_CHINESE  ),
    MakeSubtypeEntry(APA_DISPLAY_C,  DSP  , APA_SIMP_CHINESE  ),

    MakeSubtypeEntry(APA_DISPLAY_D,  DSP  , APA_JAPAN         ),  // 0x2D
    MakeSubtypeEntry(APA_DISPLAY_D,  DSP  , APA_KOREA         ),
    MakeSubtypeEntry(APA_DISPLAY_D,  DSP  , APA_TRAD_CHINESE  ),
    MakeSubtypeEntry(APA_DISPLAY_D,  DSP  , APA_SIMP_CHINESE  ),

    MakeSubtypeEntry(PRINTER_4689_THERMAL , PRN  , 4689_301   ),  // 0x32

    MakeSubtypeEntry(PRINTER_4     , PRN  , M4A               ),  // 0x34

    MakeSubtypeEntry(PRINTER_4610  , PRN  , 4610_TI1_TI2      ),  // 0x35
    MakeSubtypeEntry(PRINTER_4610  , PRN  , 4610_TI3_TI4      ),
    MakeSubtypeEntry(PRINTER_4610  , PRN  , 4610_TI5          ),
    MakeSubtypeEntry(PRINTER_4610  , PRN  , 4610_SST          ),
    MakeSubtypeEntry(PRINTER_4610  , PRN  , 4610_TI3_TI4_8MEG ),
    MakeSubtypeEntry(PRINTER_4610  , PRN  , 4610_SST_8MEG     ),
    MakeSubtypeEntry(PRINTER_4610  , PRN  , MACARENA          ),
    MakeSubtypeEntry(PRINTER_4610  , PRN  , 4610_SST_2MEG     ),

    MakeSubtypeEntry(PRINTER_4689  , PRN  , 4689_1            ),  // 0x37
    MakeSubtypeEntry(PRINTER_4689  , PRN  , 4689_2            ),

    MakeSubtypeEntry(SCANNER_A     , SCAN , 4612              ),         // @P1A
    MakeSubtypeEntry(SCANNER_A     , SCAN , 4687              ),  // 0x4A
    MakeSubtypeEntry(SCANNER_A     , SCAN , 4686_1_2          ),
    MakeSubtypeEntry(SCANNER_A     , SCAN , 4686_3_4          ),
    MakeSubtypeEntry(SCANNER_A     , SCAN , 4696              ),
    MakeSubtypeEntry(SCANNER_A     , SCAN , 4697              ),
    MakeSubtypeEntry(SCANNER_A     , SCAN , 4698              ),
    MakeSubtypeEntry(SCANNER_A     , SCAN , USB_OEM           ),

    MakeSubtypeEntry(SCANNER_B     , SCAN , HHBCR_1           ),  // 0x4B
    MakeSubtypeEntry(SCANNER_B     , SCAN , HHBCR_2           ),
    MakeSubtypeEntry(SCANNER_B     , SCAN , 1520_A02          ),
    MakeSubtypeEntry(SCANNER_B     , SCAN , USB_OEM           ),

    MakeSubtypeEntry(SCALE_A       , SCALE, 4687              ),  // 0x6A
    MakeSubtypeEntry(SCALE_A       , SCALE, 4696              ),
    MakeSubtypeEntry(SCALE_A       , SCALE, 4698              ),
    MakeSubtypeEntry(SCALE_A       , SCALE, USB_OEM           ),

    MakeSubtypeEntry(SCALE         , SCALE, 4687              ),  // 0x6E
    MakeSubtypeEntry(SCALE         , SCALE, 4696              ),
    MakeSubtypeEntry(SCALE         , SCALE, 4698              ),
    MakeSubtypeEntry(SCALE         , SCALE, USB_OEM           ),

    MakeSubtypeEntry(SYSTEM_ATTACHED_KBD, KBD, ANPOS_2        ),  // Ox81

#if defined(POSTOUCH)
    MakeSubtypeEntry(TOUCH_SCREEN  , TOUCH, 4695_A            ),         // @P1A
    MakeSubtypeEntry(TOUCH_SCREEN  , TOUCH, 4695_B            ),         // @P1A
    MakeSubtypeEntry(TOUCH_SCREEN  , TOUCH, 4695_104_M        ),  // 0x5C
    MakeSubtypeEntry(TOUCH_SCREEN  , TOUCH, 4695_95_C         ),
    MakeSubtypeEntry(TOUCH_SCREEN  , TOUCH, 4695_104_C        ),
    MakeSubtypeEntry(TOUCH_SCREEN  , TOUCH, 4695_121_C        ),
    MakeSubtypeEntry(TOUCH_SCREEN  , TOUCH, SUREPOINT_95_M    ),
    MakeSubtypeEntry(TOUCH_SCREEN  , TOUCH, SUREPOINT_95_C    ),
    MakeSubtypeEntry(TOUCH_SCREEN  , TOUCH, SUREPOINT_104_C   ),
    MakeSubtypeEntry(TOUCH_SCREEN  , TOUCH, 4820              ),

    MakeSubtypeEntry(TOUCH_SCREEN_B, TOUCH, 4695_A            ),         // @P1A
    MakeSubtypeEntry(TOUCH_SCREEN_B, TOUCH, 4695_B            ),         // @P1A
    MakeSubtypeEntry(TOUCH_SCREEN_B, TOUCH, 4695_104_M        ),  // 0x5D
    MakeSubtypeEntry(TOUCH_SCREEN_B, TOUCH, 4695_95_C         ),
    MakeSubtypeEntry(TOUCH_SCREEN_B, TOUCH, 4695_104_C        ),
    MakeSubtypeEntry(TOUCH_SCREEN_B, TOUCH, 4695_121_C        ),
    MakeSubtypeEntry(TOUCH_SCREEN_B, TOUCH, SUREPOINT_95_M    ),
    MakeSubtypeEntry(TOUCH_SCREEN_B, TOUCH, SUREPOINT_95_C    ),
    MakeSubtypeEntry(TOUCH_SCREEN_B, TOUCH, SUREPOINT_104_C   ),
    MakeSubtypeEntry(TOUCH_SCREEN_B, TOUCH, 4820              ),
#endif

    { 0, 0 }
  };


/********************************************************************/
/*    Subroutines                                                   */
/********************************************************************/


/********************************************************************/
/*                                                                  */
/*   GetTableEntry                                                  */
/*                                                                  */
/*   Given a integer value and lookup table, this returns the name  */
/*   of the constant associated with the number.  If the number     */
/*   isn't in the table, the string contains the number itself.     */
/*                                                                  */
/********************************************************************/

char *GetTableEntry(TableEntry *pLookupTable, int EntryNum)
{
    char *rc;
    static char EntryBuf[80];

    while ((pLookupTable->Value != EntryNum) && (pLookupTable->Name != 0))
      {
        pLookupTable++;
      }

    if (pLookupTable->Name)
      {
        rc = pLookupTable->Name;
      }
    else
      {
        sprintf(EntryBuf, "%-6d (0x%08X)", EntryNum, EntryNum);
        rc = EntryBuf;
      }

    return rc;
}


/*****************************************************************/
/*                                                               */
/*   DeviceStatus                                                */
/*                                                               */
/*   Prints out a message received from the POS input queue.     */
/*                                                               */
/*****************************************************************/

void DeviceStatus(POSQMSG *pMsg)
{
    if (POSM_SYS_DEVICE_ONLINE  == pMsg->msg ||
        POSM_SYS_DEVICE_OFFLINE == pMsg->msg)
      {
        fprintf(stderr,"%s - %s, %s, %s: (%s)\n",
               GetTableEntry(QMsgTable, (int) pMsg->msg),
               GetTableEntry(SlotTable, (int) (pMsg->mp1 >> 24) & 0xFF),
               GetTableEntry(PortTable, (int) (pMsg->mp1 >> 16) & 0xFF),
               GetTableEntry(TypeTable, (int) (pMsg->mp1      ) & 0xFF),
               GetTableEntry(DevTable,  (int) (pMsg->mp1 >>  8) & 0xFF));
      }
    else
      {
        fprintf(stderr,"POSQ Message from Handle %2d: %s\n",
               pMsg->mp1 & 0xFFFF,
               GetTableEntry(QMsgTable, (int) pMsg->msg));

        fprintf(stderr,"     mp1 = 0x%08X (%lu)\n", pMsg->mp1, pMsg->mp1);
        fprintf(stderr,"     mp2 = 0x%08X (%lu)\n", pMsg->mp2, pMsg->mp2);
      }

    return;
}


/********************************************************************/
/* CloseAllDevices                                                  */
/********************************************************************/

void CloseAllDevices(devicelist devInfo, int numberOfDevices)
{
    int i;

    /*
     * Search through all the devices we know about.  For each one,
     * see if it has been opened and acquired.  If so, close the
     * device.
     */
    for (i = 0; i < numberOfDevices; i++)
      {
        /*
         * If the device has been opened and acquired, close the
         * device.
         */
        if (devInfo[i].devdes > 0)
          {
            if (PosClose(devInfo[i].devdes) == -1)
              {
                fprintf(stderr, "PosIOCtl (CLOSE): errno = %s\n",
                                GetTableEntry(ErrnoTable, PosErrno()));
              }
          }
      }

  return;
}


/********************************************************************/
/* ItemLookup                                                       */
/********************************************************************/

long ItemLookup(char *sku, char *name)
{
  char buffer[80], *sz;
  int c;
  long amt;

  /*
   *  Init to 'item not found'.
   */
  amt = 0;
  strcpy(name, "ITEM NOT FOUND");

  /*
   *  Verify item record file is open.
   */
  if (item_rec_file)
    {
      /*
       * Seek to the beginning of the file.
       */
      fseek(item_rec_file, 0, SEEK_SET);

      /*
       * Read each line of the item record file.
       */
      while (fgets(buffer,sizeof(buffer),item_rec_file))
        {
          /*
           *  If the line was longer than the buffer
           *  then dispose of the excess.
           */
          if (!strchr(buffer,'\n'))
            {
              while ((c = getc(item_rec_file)) != '\n' && c != EOF)
                {
                  /* null body */
                }
            }

          /*
           *  Get the item number from the item record.
           */
          if (NULL != (sz = strtok(buffer," \t")))
            {
              /*
               *  Match?
               */
              if (!strcmp(sz,sku))
                {
                  /*
                   *  Get the price.
                   */
                  sz = strtok(NULL," \t");
                  amt = atol(sz);

                  /*
                   *  Get the description.
                   */
                  sz += (strlen(sz) + 1);
                  sz += strspn(sz," \t");
                  strncpy(name, sz, 20);
                  name[20] = '\0';
                  if (NULL != (sz = strchr(name,'\n')))
                    {
                      *sz = '\0';
                    }
                }
            }
        }
    }

  return amt;
}


/********************************************************************/
/* GetPrinterStatus                                                 */
/********************************************************************/

void GetPrinterStatus(int devdes, long *status)
{
    *status = 0;

    /*
     *  If the device has not been opened and acquired, ignore this
     *  request.
     */
    if (devdes > 0)
      {
        PosArg arg;

        /*
         * Set the printer resource to retrieve and get the information.
         */
        arg.name  = PosNprintStatus;
        arg.value = 0;

        if (PosIOCtl(devdes, POS_SYS_GET_VALUES, &arg, 1) == -1)
          {
            fprintf(stderr, "PosIOCtl (STATUS) Printer: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }

        *status = arg.value;
      }

    return;
}


/********************************************************************/
/* LockDevice                                                       */
/********************************************************************/

void LockDevice(int devdes)
{
    /*
     *  If the device has not been opened and acquired, ignore this
     *  request.
     */
    if (devdes > 0)
      {
        /*
         * Lock the device so that we don't get unexpected data.
         * Note:  some devices do not honor this request.  They
         * can still allow input to unexpectedly arrive to our
         * application.  However, the data received in the locked
         * state will have an error indication bit turned on.
         */
        if (PosIOCtl(devdes, POS_SYS_LOCK_DEVICE, NULL, 0) == -1)
          {
            fprintf(stderr, "PosIOCtl (LOCK): errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }
      }

    return;
}


/********************************************************************/
/* OpenCashDrawer                                                   */
/********************************************************************/

void OpenCashDrawer(int devdes)
{
    /*
     *  If the device has not been opened and acquired, ignore this
     *  request.
     */
    if (devdes > 0)
      {
        /*
         * Open the cash drawer till
         */
        if (PosIOCtl(devdes, POS_TILL_OPEN_TILL, NULL, 0) == -1)
          {
            fprintf(stderr, "PosIOCtl (OPEN) Cash Drawer: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }
      }

    return;
}


/********************************************************************/
/* ReadMSR                                                          */
/********************************************************************/

int ReadMSR(int devdes, char *data, int *length)
{
    int rc = FALSE;

    /*
     *  If the device has not been opened and acquired, ignore this
     *  request.
     */
    if (devdes > 0)
      {
        ssize_t readc;

        /*
         * Read the information from the MSR
         */
        if ((readc = PosRead(devdes, data, (size_t) *length)) == -1)
          {
            fprintf(stderr, "PosRead MSR: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }
        else
          {
            /*
             * If we successfully read the information, return the
             * length of the data we received to the calling function.
             */
            *length = readc;
            rc      = TRUE;
          }
      }

    return rc;
}


/********************************************************************/
/* ReadNVRAM                                                        */
/********************************************************************/

void ReadNVRAM(long *cash, long *credit, long *tax)
{
    /*
     *  If the device has not been opened and acquired, ignore this
     *  request.
     */
    if (devInfo[RAM].devdes != -1)
      {
        PosArg     set_cursor;
        nvram_data read_nvram;

        /*
         * Set the cursor position to the place we want to read the
         * data from.
         */
        set_cursor.name = PosNnvramCursor;
        set_cursor.value = (long) 0;

        if (PosIOCtl(devInfo[RAM].devdes, POS_SYS_SET_VALUES, &set_cursor, 1) == -1)
          {
            fprintf(stderr, "PosIOCtl (SET) NVRAM: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }

        /*
         * Read the data
         */
        if ((PosRead(devInfo[RAM].devdes, (char *) &read_nvram,
                     (size_t) (sizeof(read_nvram)))) == -1)
          {
            fprintf(stderr, "PosRead NVRAM: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
            *cash   = 0;
            *credit = 0;
            *tax    = 0;
          }
        else
          {
            /*
             * If we successfully read the information, return the
             * values we read.
             */
            *cash   = read_nvram.cash;
            *credit = read_nvram.credit;
            *tax    = read_nvram.tax;
          }
      }
    else
      {
        /*
         * If the device was not opened, set our values to zero.
         */
        *cash   = 0;
        *credit = 0;
        *tax    = 0;
      }

    return;
}


/********************************************************************/
/* ReadMicr                                                         */
/********************************************************************/

int ReadMicr(char *data, long *length)
{
    int rc = FALSE;

    MicrInfoRequested = FALSE;

    /*
     *  If the device has not been opened and acquired, ignore this
     *  request.
     */
    if (devInfo[PRT].devdes != -1)
      {
        ssize_t readc;

        /*
         * Read the MICR information from the printer
         */
        if ((readc = PosRead(devInfo[PRT].devdes, data, (size_t) *length)) == -1)
          {
            fprintf(stderr, "PosRead Printer: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }
        else
          {
            /*
             * If we successfully read the information, return the
             * length of the data we received to the calling function.
             */
            *length = readc;
            rc      = TRUE;
          }
      }

    return rc;
}


/********************************************************************/
/* ReadScanner                                                      */
/********************************************************************/

int ReadScanner(int devdes, char *data, int *length)
{
    int rc = FALSE;

    /*
     *  If the device has not been opened and acquired, ignore this
     *  request.
     */
    if (devdes > 0)
      {
        ssize_t readc;

        /*
         * Read the information from the scanner
         */
        if ((readc = PosRead(devdes, data, (size_t) *length)) == -1)
          {
            fprintf(stderr, "PosRead Scanner: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }
        else
          {
            /*
             * If we successfully read the information, return the
             * length of the data we received to the calling function.
             */
            *length = readc;
            rc      = TRUE;
          }
      }

    return rc;
}


/********************************************************************/
/* ReleaseAll                                                       */
/********************************************************************/

void ReleaseAll(devicelist devInfo, int numberOfDevices)
{
    int i;

    /*
     * Search through all the devices we know about.  For each one,
     * see if it has been opened and acquired.  If so, release the
     * device so other applications can use these devices.
     */
    for (i = 0; i < numberOfDevices; i++)
      {
        /*
         * If the device has been opened and acquired, release the
         * device.
         */
        if (devInfo[i].devdes > 0)
          {
            if (PosIOCtl(devInfo[i].devdes, POS_SYS_RELEASE_DEVICE,
                         NULL, 0) == -1)
              {
                fprintf(stderr, "PosIOCtl (Release Device): errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
              }
          }
      }

    return;
}


/********************************************************************/
/* SetCRPrinting                                                    */
/********************************************************************/

int SetCRPrinting(void)
{
    int rc = -1;

    /*
     *  If the device has not been opened and acquired, ignore this
     *  request.
     */
    if (devInfo[PRT].devdes != -1)
      {
        /*
         * Disable writing to the DI station.  This allows any data
         * we write after this call to be sent immediately to the
         * SJ station.  Note: if there is a document in the DI
         * station, an error will occur.
         */
        if ((rc = PosIOCtl(devInfo[PRT].devdes, POS_PRN_DISABLE_DI_PRINTING, 0, 0)) == -1)
          {
            /*  Error enabling CR printing */
            fprintf(stderr, "PosIOCtl (Enable CR Station): errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }
        else
          {
            PosArg resource;

            /*
             * Since we successfully disabled DI printing, we can now
             * set the station to allow writing to the CR station.
             */
            resource.name  = PosNprintStation;
            resource.value = PosPRINT_STATION_CR;

            if ((rc = PosIOCtl(devInfo[PRT].devdes, POS_SYS_SET_VALUES, &resource, 1)) == -1)
              {
                /*  Error setting resource */
                fprintf(stderr, "PosIOCtl (Set CR Station): errno = %s\n",
                                GetTableEntry(ErrnoTable, PosErrno()));
              }
          }
      }

    return rc;
}


/********************************************************************/
/* SetDIPrinting                                                    */
/********************************************************************/

int SetDIPrinting(void)
{
    int rc = -1;

    /*
     *  If the device has not been opened and acquired, ignore this
     *  request.
     */
    if (devInfo[PRT].devdes != -1)
      {
        /*
         * Enable writing to the DI station.  This allows any data
         * we write after this call to be sent immediately to the
         * DI station.
         */
        if ((rc = PosIOCtl(devInfo[PRT].devdes, POS_PRN_ENABLE_DI_PRINTING, 0, 0)) == -1)
          {
            /*  Error enabling DI printing */
            fprintf(stderr, "PosIOCtl (Enable DI Station): errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }
        else
          {
            PosArg resource;

            /*
             * Since we successfully enabled DI printing, we can now
             * set the station to allow writing to the DI station.
             */
            resource.name  = PosNprintStation;
            resource.value = PosPRINT_STATION_DI;

            if ((rc = PosIOCtl(devInfo[PRT].devdes, POS_SYS_SET_VALUES, &resource, 1)) == -1)
              {
                /*  Error setting resource */
                fprintf(stderr, "PosIOCtl (Set DI Station): errno = %s\n",
                                GetTableEntry(ErrnoTable, PosErrno()));
              }
          }
      }

    return rc;
}


/********************************************************************/
/* SetSJPrinting                                                    */
/********************************************************************/

int SetSJPrinting(void)
{
    int rc = -1;

    /*
     *  If the device has not been opened and acquired, ignore this
     *  request.
     */
    if (devInfo[PRT].devdes != -1)
      {
        /*
         * Disable writing to the DI station.  This allows any data
         * we write after this call to be sent immediately to the
         * SJ station.  Note: if there is a document in the DI
         * station, an error will occur.
         */
        if ((rc = PosIOCtl(devInfo[PRT].devdes, POS_PRN_DISABLE_DI_PRINTING, 0, 0)) == -1)
          {
            /*  Error enabling SJ printing */
            fprintf(stderr, "PosIOCtl (Enable SJ Station): errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }
        else
          {
            PosArg resource;

            /*
             * Since we successfully disabled DI printing, we can now
             * set the station to allow writing to the SJ station.
             */
            resource.name  = PosNprintStation;
            resource.value = PosPRINT_STATION_SJ;

            if ((rc = PosIOCtl(devInfo[PRT].devdes, POS_SYS_SET_VALUES, &resource, 1)) == -1)
              {
                /*  Error setting resource */
                fprintf(stderr, "PosIOCtl (Set SJ Station): errno = %s\n",
                                GetTableEntry(ErrnoTable, PosErrno()));
              }
          }
      }

    return rc;
}


/********************************************************************/
/* SetCheckoutNumpadStyle                                           */
/********************************************************************/

void SetCheckoutNumpadStyle(int devdes)
{
    /*
     *  If the device has not been opened and acquired, ignore this
     *  request.
     */
    if (devdes > 0)
      {
        PosArg resource;

        /*
         * The original checkout keyboards were shipped with the numpad
         * in the touch tone position.  Tell the device handler that
         * we want to honor this style.
         */
        resource.name  = PosNnumpadStyle;
        resource.value = PosTOUCHTONE_POS;

        /*
         * Change the keyboard numpad style
         */
        if (PosIOCtl(devdes, POS_SYS_SET_VALUES, &resource, 1) == -1)
          {
            fprintf(stderr, "PosIOCtl (SET_VALUES) Keyboard errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }
      }

    return;
}


/********************************************************************/
/* SoundTone                                                        */
/********************************************************************/

void SoundTone()
{
    /*
     *  If a tone device (keyboard or touch screen) has not been
     *  opened and acquired, ignore this request.
     */
    if (devInfo[KBD].devdes > 0)
      {
        /*
         * Sound the keyboard tone.  The duration and frequency
         * are set through the devices resources.  We can override
         * the values in our resource file.
         */
        if (PosIOCtl(devInfo[KBD].devdes, POS_KBD_SOUND_TONE, NULL, 0) == -1)
          {
            fprintf(stderr, "PosIOCtl (SOUND) Keyboard errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }
      }

#if defined(POSTOUCH)
    else if (devInfo[TCH].devdes > 0)
      {
        /*
         * Sound the touch screen tone.  The duration and frequency
         * are set through the devices resources.  We can override
         * the values in our resource file.
         */
        if (PosIOCtl(devInfo[TCH].devdes, POS_TOUCH_SOUND_TONE, NULL, 0) == -1)
          {
            fprintf(stderr, "PosIOCtl (SOUND) Touch errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }
      }
#endif

    return;
}


/********************************************************************/
/* UnlockDevice                                                     */
/********************************************************************/

void UnlockDevice(int devdes)
{
    /*
     *  If the device has not been opened and acquired, ignore this
     *  request.
     */
    if (devdes > 0)
      {
        /*
         * Unlock the device
         */
        if (PosIOCtl(devdes, POS_SYS_UNLOCK_DEVICE, NULL, 0) == -1)
          {
            fprintf(stderr, "PosIOCtl (UNLOCK): errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }
      }

    return;
}


/********************************************************************/
/* UpdateNVRAM                                                      */
/********************************************************************/

void UpdateNVRAM(int devdes, long cash, long credit, long tax)
{
    /*
     *  If the device has not been opened and acquired, ignore this
     *  request.
     */
    if (devdes > 0)
      {
        PosArg set_cursor;
        ssize_t written;
        nvram_data rw_nvram;

        /*
         * Set the cursor position to the place we want to read the
         * data from.
         */
        set_cursor.name = PosNnvramCursor;
        set_cursor.value = (long) 0;

        if (PosIOCtl(devdes, POS_SYS_SET_VALUES, &set_cursor, 1) == -1)
          {
            fprintf(stderr, "PosIOCtl (SET) NVRAM: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }

        /*
         * Read the data
         */
        if ((PosRead(devdes, (char *)&rw_nvram,
                     (size_t) (sizeof(rw_nvram)))) == -1)
          {
            fprintf(stderr, "PosRead NVRAM: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }

        /*
         * Add the new amounts to the data we read in.
         */
        rw_nvram.cash   += cash;
        rw_nvram.credit += credit;
        rw_nvram.tax    += tax;

        /*
         * Set the cursor position to where we want to write the
         * data in NVRAM
         */
        set_cursor.value = (long) 0;

        if (PosIOCtl(devdes, POS_SYS_SET_VALUES, &set_cursor, 1) == -1)
          {
            fprintf(stderr, "PosIOCtl (SET) NVRAM: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }

        /*
         * Write the data to NVRAM.  This call will block our
         * program until the data has successfully been written
         * to NVRAM.
         */
        if ((written = PosWrite(devdes, (char *)&rw_nvram,
                               (size_t) (sizeof(rw_nvram)))) == -1)
          {
            fprintf(stderr, "PosWrite NVRAM: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }
        else if (written != (ssize_t) (sizeof(rw_nvram)))
          {
            fprintf(stderr, "PosWrite NVRAM: wrote %d of %d bytes\n",
                            written, (sizeof(long)));
           }
       }

    return;
}


/********************************************************************/
/* ClearDisplay                                                     */
/********************************************************************/

void ClearDisplay(int devdes)
{
    /*
     *  If the device has not been opened and acquired, ignore this
     *  request.
     */
    if (devdes > 0)
      {
        /*
         * Clear the entire display
         */
        if (PosIOCtl(devdes, POS_DSP_CLEAR_SCREEN, NULL, 0) == -1)
          {
            fprintf(stderr, "PosIOCtl (CLEAR_SCREEN): errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }
      }

#if defined(POSTOUCH)
    /*
     * Clear the touch screen mirror.
     */
    WriteTouchMirror(0, "                                        ", 40);
#endif

    return;
}


/********************************************************************/
/* DisplayFullScreen                                                */
/********************************************************************/

void DisplayFullScreen(int devdes, char *message, int cursor)
{
    PosArg  set_cursor;
    int     length;
    ssize_t written;

    set_cursor.name  = PosNdisplayCursor;
    set_cursor.value = (long) cursor;

    /*
     * Calculate how long the character string is.  The maximum
     * length is 40.
     */
    if ((length = strlen(message)) >= 40)
      {
        length = 40;
      }

    /*
     *  If the display device has been opened and acquired,
     *  clear the display, position the cursor, and write the
     *  message to the display.
     */
    if (devdes > 0)
      {
        ClearDisplay(devdes);

        if (PosIOCtl(devdes, POS_SYS_SET_VALUES, &set_cursor, 1) == -1)
          {
            fprintf(stderr, "PosIOCtl (SET) cursor: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }

        if ((written = PosWrite(devdes, message, (size_t) length)) == -1)
          {
            fprintf(stderr, "PosWrite Display: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }
        else if (written != (ssize_t) length)
          {
            fprintf(stderr, "PosWrite Display: wrote %d of %d bytes\n",
                            written, length);
          }
      }

#if defined(POSTOUCH)
    /*
     *  Write the message to the touch screen mirror.
     */
    WriteTouchMirror(0, "                                        ", 40);
    WriteTouchMirror(set_cursor.value, message, length);
#endif

    return;
}


/********************************************************************/
/* DisplayString                                                    */
/********************************************************************/

void DisplayString(int devdes, char *message, int cursor)
{
    PosArg  set_cursor;
    char    buffer[21];
    int     length;
    ssize_t written;

    /*
     * Calculate how long the character string is.  The maximum
     * length is 20.
     */
    if ((length = strlen(message)) >= 20)
      {
        memcpy(buffer, message, 20);
      }
    else
      {
        memcpy(buffer, message, length);
        memset(buffer + length, ' ', 20 - length);
      }

    set_cursor.name  = PosNdisplayCursor;
    set_cursor.value = cursor >= 20 ? (long) 20 : (long) 0;

    /*
     *  If the display device has been opened and acquired,
     *  position the cursor and write the message to the display.
     */
    if (devdes > 0)
      {
        if (PosIOCtl(devdes, POS_SYS_SET_VALUES, &set_cursor, 1) == -1)
          {
            fprintf(stderr, "PosIOCtl (SET) cursor: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }

        if ((written = PosWrite(devdes, buffer, (size_t) 20)) == -1)
          {
            fprintf(stderr, "PosWrite Display: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }
        else if (written != (ssize_t) 20)
          {
            fprintf(stderr, "PosWrite Display: wrote %d of 20 bytes\n",
                            written);
          }
      }

#if defined(POSTOUCH)
    /*
     *  Write the message to the touch screen mirror.
     */
    WriteTouchMirror(set_cursor.value, buffer, 20);
#endif

    return;
}


/********************************************************************/
/* DisplayItemRight                                                 */
/********************************************************************/

void DisplayItemRight(int devdes, char *message)
{
    PosArg  set_cursor;
    int     length;
    ssize_t written;

    /*
     * Calculate how long the character string is.  The maximum
     * length is 40.
     */
    if ((length = strlen(message)) >= 40)
      {
        length = 40;
      }

    if (length > 0)
      {
        /*
         * Set the cursor to where we want it
         */
        set_cursor.name  = PosNdisplayCursor;
        set_cursor.value = (long) (40 - length);

        /*
         *  If the display device has been opened and acquired,
         *  position the cursor and write the message to the display.
         */
        if (devdes > 0)
          {
            if (PosIOCtl(devdes, POS_SYS_SET_VALUES, &set_cursor, 1) == -1)
              {
                fprintf(stderr, "PosIOCtl (SET) cursor: errno = %s\n",
                                GetTableEntry(ErrnoTable, PosErrno()));
              }

            if ((written = PosWrite(devdes, message, (size_t) length)) == -1)
              {
                fprintf(stderr, "PosWrite Display: errno = %s\n",
                                GetTableEntry(ErrnoTable, PosErrno()));
              }
            else if (written != (ssize_t) length)
              {
                fprintf(stderr, "PosWrite Display: wrote %d of %d bytes\n",

                                written, length);
              }
          }

#if defined(POSTOUCH)
        /*
         *  Write the message to the touch screen mirror.
         */
        WriteTouchMirror(set_cursor.value, message, length);
#endif
      }

    return;
}


/********************************************************************/
/* ltoa                                                             */
/********************************************************************/
char *ltoa(long l, char *a, size_t a_size)
{
  char ach[20], *pch = ach;

  sprintf(ach,"%ld",l);
  if (strlen(ach) + 1 > a_size)
    {
      pch += strlen(ach) + 1 - a_size;
    }
  strcpy(a,ach);
  return a;
}


/********************************************************************/
/* DisplayAmount                                                    */
/********************************************************************/

void DisplayAmount(int devdes, long amt)
{
    char buffer[21];
    char msg[21];
    char *pAmount = ltoa(amt, msg, 10);
    int length    = strlen(msg);

    /*
     * If the length is equal to one, then use only the last position
     * Otherwise, if the length is equal to two, use set the dollar
     * amount to zero and use only the cents portion.  Finally, if
     * the length is greater than two, use the data as is and add in
     * a period to mark the dollar/cents seperator.
     */
    if (length == 1)
      {
        buffer[0] = ' ';
        buffer[1] = ' ';
        buffer[2] = ' ';
        buffer[3] = ' ';
        buffer[4] = ' ';
        buffer[5] = ' ';
        buffer[6] = '0';
        buffer[7] = '.';
        buffer[8] = '0';
        buffer[9] = msg[0];
        buffer[10] = '\0';
      }
    else if (length == 2)
      {
        buffer[0] = ' ';
        buffer[1] = ' ';
        buffer[2] = ' ';
        buffer[3] = ' ';
        buffer[4] = ' ';
        buffer[5] = ' ';
        buffer[6] = '0';
        buffer[7] = '.';
        buffer[8] = msg[0];
        buffer[9] = msg[1];
        buffer[10] = '\0';
      }
    else
      {
        memcpy(buffer, msg, length - 2);
        buffer[length - 2] = '.';
        memcpy(&buffer[length - 1], &msg[length - 2], 2);
        buffer[length + 1] = '\0';
      }

    /*
     * Display the amount on the screen
     */
    DisplayItemRight(devdes, buffer);

    return;
}


/********************************************************************/
/* WritePrinterClose                                                */
/********************************************************************/

void WritePrinterClose(char *item, long amt)
{
    /*
     *  If the device has not been opened and acquired, ignore this
     *  request.
     */
    if (devInfo[PRT].devdes != -1)
      {
        char buffer[40];

        /*
         *  Format the data to be printed
         */
        sprintf(buffer, "%-25s%10.2f\r\n", item, (double) amt / (double) 100);
        if (devInfo[PRT].dev_number != PosDEVICE_PRINTER_4610)
          {
            WriteToSJstation(buffer);
          }
        else
          {
            WriteToCRstation(buffer);
          }
      }

    return;
}


/********************************************************************/
/* WriteToCRstation                                                 */
/********************************************************************/

void WriteToCRstation(char *line)
{
    /*
     *  If the device has not been opened and acquired, ignore this
     *  request.
     */
    if (devInfo[PRT].devdes != -1)
      {
        PosArg  set_station;
        ssize_t written;
        size_t  length = (size_t) strlen(line);

        /*
         *  Set the print station to the CR station.  All subsequent
         *  writes will be to this station until it is changed.
         */
        set_station.name  = PosNprintStation;
        set_station.value = PosPRINT_STATION_CR;

        if (PosIOCtl(devInfo[PRT].devdes, POS_SYS_SET_VALUES, &set_station, 1) == -1)
          {
            fprintf(stderr, "PosIOCtl (SET) Printer: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }

        /*
         * Write the data to the CR station
         */
        if ((written = PosWrite(devInfo[PRT].devdes, line, length)) == -1)
          {
            fprintf(stderr, "PosWrite Printer: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }
        else if (written != (ssize_t) length)
          {
            fprintf(stderr, "PosWrite Printer: wrote %d of %d bytes\n",
                            written, length);
          }
      }

    return;
}


/********************************************************************/
/* WriteToCRandSJstations                                           */
/********************************************************************/

void WriteToCRandSJstations(char *buffer)
{
    /*
     *  If the printer device has not been opened and acquired, ignore this
     *  request.
     */
    if (devInfo[PRT].devdes != -1)
      {
        PosArg  set_station;
        ssize_t written;
        size_t  length = (size_t) strlen(buffer);

        /*
         *  Set the print station to the CR and SJ stations.  All subsequent
         *  writes will be to this station until it is changed.
         */
        set_station.name  = PosNprintStation;
        set_station.value = PosPRINT_STATION_CR + PosPRINT_STATION_SJ;

        if (PosIOCtl(devInfo[PRT].devdes, POS_SYS_SET_VALUES, &set_station, 1) == -1)
          {
            fprintf(stderr, "PosIOCtl (SET) Printer: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }

        /*
         * Write the data to the CR and SJ stations
         */
        if ((written = PosWrite(devInfo[PRT].devdes, buffer, length)) == -1)
          {
            fprintf(stderr, "PosWrite Printer: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }
        else if (written != (ssize_t) length)
          {
            fprintf(stderr, "PosWrite Printer: wrote %d of %d bytes\n",
                            written, length);
          }
      }

    return;
}


/********************************************************************/
/* WriteToDIstation                                                 */
/********************************************************************/

void WriteToDIstation(char *buffer, size_t length)
{
    /*
     *  If the device has not been opened and acquired, ignore this
     *  request.
     */
    if (devInfo[PRT].devdes != -1)
      {
        PosArg  set_station;
        ssize_t written;

        /*
         *  Set the print station to the DI station.  All subsequent
         *  writes will be to this station until it is changed.
         */
        set_station.name  = PosNprintStation;
        set_station.value = PosPRINT_STATION_DI;

        if (PosIOCtl(devInfo[PRT].devdes, POS_SYS_SET_VALUES, &set_station, 1) == -1)
          {
            fprintf(stderr, "PosIOCtl (SET) Printer: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }

        /*
         * Write the data to the DI station
         */
        if ((written = PosWrite(devInfo[PRT].devdes, buffer, length)) == -1)
          {
            fprintf(stderr, "PosWrite Printer: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }
        else if (written != (ssize_t) length)
          {
            fprintf(stderr, "PosWrite Printer: wrote %d of %d bytes\n",
                            written, length);
          }
      }

    return;
}


/********************************************************************/
/* WritePrinterDate                                                 */
/********************************************************************/

void WritePrinterDate(int tr_num)
{
    /*
     *  If the device has not been opened and acquired, ignore this
     *  request.
     */
    if (devInfo[PRT].devdes != -1)
      {
        time_t ltime = time((time_t *) 0);
        char   buffer[41];
        char   dateBuffer[41];

        /*
         * Format the current time into a printable string and remove
         * the trailing new-line character.
         */
        strcpy(dateBuffer, ctime(&ltime));
        dateBuffer[strlen(dateBuffer) - 1] = '\0';

        /*
         * Format the data to be printed
         */
        sprintf(buffer, "\r\n%-25s      %04d\r\n", dateBuffer, tr_num);

        /*
         * Write the data to both the CR and the SJ stations
         */
        if (devInfo[PRT].dev_number != PosDEVICE_PRINTER_4610)
          {
            WriteToCRandSJstations(buffer);
          }
        else
          {
            WriteToCRstation(buffer);
          }

      }

    return;
}


/********************************************************************/
/* WriteToSJstation                                                 */
/********************************************************************/

void WriteToSJstation(char *line)
{
    /*
     *  If the device has not been opened and acquired, ignore this
     *  request.
     */
    if (devInfo[PRT].devdes != -1)
      {
        PosArg  set_station;
        ssize_t written;
        size_t  length = (size_t) strlen(line);

        /*
         *  Set the print station to the SJ station.  All subsequent
         *  writes will be to this station until it is changed.
         */
        set_station.name  = PosNprintStation;
        set_station.value = PosPRINT_STATION_SJ;

        if (PosIOCtl(devInfo[PRT].devdes, POS_SYS_SET_VALUES, &set_station, 1) == -1)
          {
            fprintf(stderr, "PosIOCtl (SET) Printer: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }

        /*
         * Write the data to the SJ station
         */
        if ((written = PosWrite(devInfo[PRT].devdes, line, length)) == -1)
          {
            fprintf(stderr, "PosWrite Printer: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }
        else if (written != (ssize_t) length)
          {
            fprintf(stderr, "PosWrite Printer: wrote %d of %d bytes\n",
                            written, length);
          }
      }

    return;
}


/********************************************************************/
/* WriteFeatures                                                    */
/********************************************************************/

int WriteFeatures(void)
{
    int rc = 0;

    if (devInfo[PRT].devdes != -1 && pos_state == Signed_Off_State)
      {
        FILE *fd;

        if ((fd = fopen("features.aip", "r")) != (FILE *) NULL)
          {
            char buffer[256];
            char prnBuffer[256];

            pos_state = Print_Features_State;

            DisplayFullScreen(devInfo[DSP].devdes, "Printing Features", 0);

            while (fgets(buffer, 255, fd) != (char *) NULL)
              {
                if (buffer[0] == '#')
                  {
                    continue;
                  }

                buffer[strlen(buffer) - 1] = '\0';
                buffer[38] = '\0';

                sprintf(prnBuffer, "%s\r\n", buffer);

                WriteToCRstation(prnBuffer);
              }

            WriteToCRstation(chase_print);

            fclose(fd);

          }
        else
          {
            rc = -1;
            SoundTone();
          }
      }
    else
      {
        rc = -1;
        SoundTone();
      }

    return rc;
}


/********************************************************************/
/* WriteTransactionLog                                              */
/********************************************************************/

int WriteTransactionLog(transactionRecord *tran)
{
    FILE *fd;
    int  rc = 0;

    if ((fd = fopen(tran->logFileName, "a+")) != (FILE *) NULL)
      {
        char signOffBuf[50];
        time_t ltime = time((time_t *) 0);

        sprintf(signOffBuf, " SIGN-OFF %05d %.24s",
                            tran->transact_num, ctime(&ltime));

        signOffBuf[strlen(signOffBuf)] = '\0';

        fprintf(fd, "%-46s%-120s TAX%8ld CASH%8ld CREDIT%8ld%-s\n",
                    tran->signOnBuf,
                    tran->item_list,
                    tran->log_tax_amount,
                    tran->log_cash_amount,
                    tran->log_credit_amount,
                    signOffBuf);

        fclose(fd);

        tran->signOnBuf[0]      = '\0';
        tran->item_list[0]      = '\0';
        tran->log_tax_amount    = 0;
        tran->log_cash_amount   = 0;
        tran->log_credit_amount = 0;
      }
    else
      {
        fprintf(stderr, "fopen \"%s\" failed\n", tran->logFileName);

        rc = -1;
      }

    return rc;
}


/********************************************************************/
/* ZeroNVRAM                                                        */
/********************************************************************/

void ZeroNVRAM(void)
{
    /*
     *  If the device has not been opened and acquired, ignore this
     *  request.
     */
    if (devInfo[RAM].devdes > 0)
      {
        PosArg     set_cursor;
        ssize_t    written;
        nvram_data zero_nvram;

        /*
         * Set the cursor position to where we want to write the
         * data in NVRAM
         */
        set_cursor.name  = PosNnvramCursor;
        set_cursor.value = (long) 0;

        if (PosIOCtl(devInfo[RAM].devdes, POS_SYS_SET_VALUES, &set_cursor, 1) == -1)
          {
            fprintf(stderr, "PosIOCtl (SET) NVRAM: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }

        /*
         * Set the initial data to zero.
         */
        memset(&zero_nvram, 0, sizeof(zero_nvram));

        /*
         * Write the data to NVRAM.  This call will block our
         * program until the data has successfully been written
         * to NVRAM.
         */
        if ((written = PosWrite(devInfo[RAM].devdes, (char *) &zero_nvram,
                                (size_t) (sizeof(zero_nvram)))) == -1)
          {
            fprintf(stderr, "PosWrite NVRAM: errno = %s\n",
                            GetTableEntry(ErrnoTable, PosErrno()));
          }
        else if (written != (ssize_t) (sizeof(zero_nvram)))
          {
            fprintf(stderr, "PosWrite NVRAM: wrote %d of %d bytes\n",
                            written, sizeof(zero_nvram));
          }
      }

    return;
}


/*****************************************************************************/
/*                                                                           */
/*  Function Name:   OpenAndAcquire                                          */
/*                                                                           */
/*  Description:                                                             */
/*                                                                           */
/*     This subroutine opens a device and then acquires exclusive usage      */
/*     of it.                                                                */
/*                                                                           */
/*****************************************************************************/

int OpenAndAcquire(int devdes, char *devName, char *className,
                   PosArgPtr args, int nargs)
{
    int rcdev = devdes;

    /*
     *  If the device has been opened and acquired, ignore this
     *  request.
     */
    if (devdes == -1 )
      {
        /*
         * Try to open the device
         */
        if ((rcdev = PosOpen(devName, className, args, nargs)) != -1)
          {
            if (PosIOCtl(rcdev, POS_SYS_ACQUIRE_DEVICE, NULL, 0) != 0)
              {
                fprintf(stderr, "Error acquiring %s: errno = %s\n",
                                className, GetTableEntry(ErrnoTable, PosErrno()));

                /*
                 * Close the device since we could not acquire use of it.
                 */
                PosClose(rcdev);
              }
          }
        else
          {
            fprintf(stderr, "Error opening %s: errno = %s\n",
                            className, GetTableEntry(ErrnoTable, PosErrno()));
          }
      }

    return rcdev;
}


/*****************************************************************************/
/*                                                                           */
/*  Function Name:   CloseDevice                                             */
/*                                                                           */
/*  Description:                                                             */
/*                                                                           */
/*     This subroutine closes the device for which an offline message has    */
/*     been received.                                                        */
/*                                                                           */
/*****************************************************************************/

void CloseDevice(POSQMSG *pMsg)
{
    int     DevType   = 0;                 /* Device type from ONLINE msg */
    int     DevNumber = 0;                 /* Device address from ONLINE msg */

    DevNumber = (int) (pMsg->mp1 >> 8) & 0x00FF;
    DevType   = (int) pMsg->mp1 & 0x00FF;

    switch (DevType)
      {
        /**********************************************************/
        /*                     Display                            */
        /**********************************************************/
        case PosTYPE_DISPLAY:
          {
            if (devInfo[DSP].dev_number == DevNumber) 
              {
                if (PosClose(devInfo[DSP].devdes) == -1)
                  {
                    fprintf(stderr, "PosClose %s: errno = %s\n",
                                    devInfo[DSP].cls_name,
                                    GetTableEntry(ErrnoTable, PosErrno()));
                  }

                devInfo[DSP].devdes = -1;
              }

            break;
          }

        /**********************************************************/
        /*                     Printer                            */
        /**********************************************************/
        case PosTYPE_PRINTER:
          {
            if (devInfo[PRT].dev_number == DevNumber)
              {
                if (PosClose(devInfo[PRT].devdes) == -1)
                  {
                    fprintf(stderr, "PosClose %s: errno = %s\n",
                            devInfo[DSP].cls_name,
                            GetTableEntry(ErrnoTable, PosErrno()));
                  }
                devInfo[PRT].devdes = -1;
              }

            break;
          }

        /**********************************************************/
        /*                 Bar Code Reader                        */
        /**********************************************************/
        case PosTYPE_SCANNER:
          {
            if (DevNumber == devInfo[SCN].dev_number) 
              {
                if (PosClose(devInfo[SCN].devdes) == -1)
                  {
                    fprintf(stderr, "PosClose %s: errno = %s\n",
                                    devInfo[SCN].cls_name,
                                    GetTableEntry(ErrnoTable, PosErrno()));
                  }

                devInfo[SCN].devdes = -1;
              }

            break;
          }

        /**********************************************************/
        /*                       Scale                            */
        /**********************************************************/
        case PosTYPE_SCALE:
          {
            if (PosClose(devInfo[SCL].devdes) == -1)
              {
                fprintf(stderr, "PosClose %s: errno = %s\n",
                        devInfo[SCL].cls_name,
                        GetTableEntry(ErrnoTable, PosErrno()));
              }

            devInfo[SCL].devdes = -1;

            break;
          }

        /**********************************************************/
        /*                 Magnetic Stripe Reader                 */
        /**********************************************************/
        case PosTYPE_MSR:
          {
            if (DevNumber == devInfo[MSR].dev_number) 
              {
                if (PosClose(devInfo[MSR].devdes) == -1)
                  {
                    fprintf(stderr, "PosClose %s: errno = %s\n",
                            devInfo[MSR].cls_name,
                            GetTableEntry(ErrnoTable, PosErrno()));
                  }

                devInfo[MSR].devdes = -1;
              }

            break;
          }

        /**********************************************************/
        /*                      Keyboard                          */
        /**********************************************************/
        case PosTYPE_KEYBOARD:
          {
            if (DevNumber == devInfo[KBD].dev_number) 
              {
                if (PosClose(devInfo[KBD].devdes) == -1)
                  {
                    fprintf(stderr, "PosClose %s: errno = %s\n",
                                    devInfo[KBD].cls_name,
                                    GetTableEntry(ErrnoTable, PosErrno()));
                  }

                devInfo[KBD].devdes = -1;
              }

            break;
          }

        /**********************************************************/
        /*                   Cash Drawer (Till)                   */
        /**********************************************************/
        case PosTYPE_CASH_DRAWER:
          {
            if (DevNumber == devInfo[TIL].dev_number) 
              {
                if (PosClose(devInfo[TIL].devdes) == -1)
                  {
                    fprintf(stderr, "PosClose %s: errno = %s\n",
                            devInfo[TIL].cls_name,
                            GetTableEntry(ErrnoTable, PosErrno()));
                  }

                devInfo[TIL].devdes = -1;
              }

            break;
          }

        /**********************************************************/
        /*                         NVRAM                          */
        /**********************************************************/
        case PosTYPE_NVRAM:
          {
            if (DevNumber == devInfo[RAM].dev_number)
              {
                if (PosClose(devInfo[RAM].devdes) == -1)
                  {
                    fprintf(stderr, "PosClose %s: errno = %s\n",
                                    devInfo[RAM].cls_name,
                                    GetTableEntry(ErrnoTable, PosErrno()));
                  }

                devInfo[RAM].devdes = -1;
              }

            break;
         }

#if defined(POSTOUCH)
        /**********************************************************/
        /*                    TOUCH SCREEN                        */
        /**********************************************************/
        case PosTYPE_TOUCH_SCREEN:
          {
            if (PosClose(devInfo[TCH].devdes) == -1)
              {
                fprintf(stderr, "PosClose %s: errno = %s\n",
                                devInfo[TCH].cls_name,
                                GetTableEntry(ErrnoTable, PosErrno()));
              }

            devInfo[TCH].devdes = -1;

            break;
         }
#endif

       default:
         {
           break;
         }
      }

    return;
}


/*****************************************************************************/
/*                                                                           */
/*  Function Name:   OpenDevice                                              */
/*                                                                           */
/*  Description:                                                             */
/*                                                                           */
/*     This subroutine opens the device for which an online message has been */
/*     received.                                                             */
/*                                                                           */
/*****************************************************************************/

void OpenDevice(POSQMSG *pMsg)
{
    PosArg  openargs[3];                   /* Arg array for PosOpen */
    int     DevType   = 0;                 /* Device type from ONLINE msg */
    int     SubType   = 0;                 /* Device subtype from ONLINE msg */
    int     DevNumber = 0;                 /* Device address from ONLINE msg */
    int     rc        = 0;                 /* Return codes    */

    PosSetArg(openargs[0], PosNslotNumber, ((int) (pMsg->mp1 >> 24) & 0x00FF));
    PosSetArg(openargs[1], PosNportNumber, ((int) (pMsg->mp1 >> 16) & 0x00FF));
    PosSetArg(openargs[2], PosNdeviceNumber, ((int) (pMsg->mp1 >> 8) & 0x00FF));
    DevNumber = (int) (pMsg->mp1 >> 8) & 0x00FF;
    DevType   = (int) pMsg->mp1 & 0x00FF;
    SubType   = (int) pMsg->mp2 & 0x00FF;

    switch (DevType)
      {
        /**********************************************************/
        /*              Non-Shopper Display                       */
        /**********************************************************/
        case PosTYPE_DISPLAY:
          {
            if ((DevNumber != PosDEVICE_SHOPPER_DISPLAY_A) &&
                (DevNumber != PosDEVICE_SHOPPER_DISPLAY_B))
              {
                if (devInfo[DSP].devdes == -1) 
                  {
                    rc = OpenAndAcquire(devInfo[DSP].devdes,
                                        devInfo[DSP].dev_name,
                                        devInfo[DSP].cls_name,
                                        openargs,
                                        PosNumber(openargs));

                    if (rc > 0)
                      {
                        devInfo[DSP].devdes = rc;
                        devInfo[DSP].dev_type = DevType;
                        devInfo[DSP].dev_subtype = SubType;
                        devInfo[DSP].dev_number = DevNumber;

                        if (welcome_displayed == TRUE)
                          {
                            /*
                             * Write the initial message to the display
                             */
                            DisplayFullScreen(devInfo[DSP].devdes, achGreeting, 0);
                          }
                      }
                  }
              }

            break;
          }

        /**********************************************************/
        /*           Printer (SBCS and non-fiscal)                */
        /**********************************************************/
        case PosTYPE_PRINTER:
          {
            if ((DevNumber == PosDEVICE_PRINTER_2) ||
                (DevNumber == PosDEVICE_PRINTER_3) ||
                (DevNumber == PosDEVICE_PRINTER_4) ||
                (DevNumber == PosDEVICE_PRINTER_4610))
              {
                if (devInfo[PRT].devdes == -1) 
                  {
                    rc = OpenAndAcquire(devInfo[PRT].devdes,
                                        devInfo[PRT].dev_name,
                                        devInfo[PRT].cls_name,
                                        openargs,
                                        PosNumber(openargs));
       
                    if (rc > 0)
                      {
                        long status;
       
                        devInfo[PRT].devdes = rc;
                        devInfo[PRT].dev_type = DevType;
                        devInfo[PRT].dev_subtype = SubType;
                        devInfo[PRT].dev_number = DevNumber;
       
                        GetPrinterStatus(devInfo[PRT].devdes, &status);
       
                        /*
                         * Is a MICR reader installed in the printer
                         */
                        if (status & PosSTATUS_MICR_INSTALLED)
                          {
                            MicrInstalled = TRUE;
                          }
                        else
                          {
                            MicrInstalled = FALSE;
                          }
                      }
                  }
              }

            break;
          }

        /**********************************************************/
        /*               Bar Code Reader/Scanner                  */
        /**********************************************************/
        case PosTYPE_SCANNER:
          {
            if (devInfo[SCN].devdes == -1) 
              {
                rc = OpenAndAcquire(devInfo[SCN].devdes,
                                    devInfo[SCN].dev_name,
                                    devInfo[SCN].cls_name,
                                    openargs,
                                    PosNumber(openargs));
       
                if (rc > 0)
                  {
                    devInfo[SCN].devdes = rc;
                    devInfo[SCN].dev_type = DevType;
                    devInfo[SCN].dev_subtype = SubType;
                    devInfo[SCN].dev_number = DevNumber;
       
                    if (PosIOCtl(devInfo[SCN].devdes,
                                 POS_SYS_UNLOCK_DEVICE, 0, 0) == -1)
                      {
                        fprintf(stderr, "Error unlocking %s: errno = %s\n",
                                        devInfo[SCN].cls_name,
                                        GetTableEntry(ErrnoTable, PosErrno()));
                      }
                  }
              }

            break;
          }

        /**********************************************************/
        /*                          Scale                         */
        /**********************************************************/
        case PosTYPE_SCALE:
          {
            if (devInfo[SCL].devdes == -1) 
              {
                rc = OpenAndAcquire(devInfo[SCL].devdes,
                                    devInfo[SCL].dev_name,
                                    devInfo[SCL].cls_name,
                                    openargs,
                                    PosNumber(openargs));
           
                if (rc > 0)
                  {
                    devInfo[SCL].devdes = rc;
                    devInfo[SCL].dev_type = DevType;
                    devInfo[SCL].dev_subtype = SubType;
                    devInfo[SCL].dev_number = DevNumber;
                  }
              }
       
            break;
          }

        /**********************************************************/
        /*                 Magnetic Stripe Reader                 */
        /**********************************************************/
        case PosTYPE_MSR:
          {
            if (devInfo[MSR].devdes == -1) 
              {
                rc = OpenAndAcquire(devInfo[MSR].devdes,
                                    devInfo[MSR].dev_name,
                                    devInfo[MSR].cls_name,
                                    openargs,
                                    PosNumber(openargs));
       
                if (rc > 0)
                  {
                    devInfo[MSR].devdes = rc;
                    devInfo[MSR].dev_type = DevType;
                    devInfo[MSR].dev_subtype = SubType;
                    devInfo[MSR].dev_number = DevNumber;
       
                    if (PosIOCtl(devInfo[MSR].devdes,
                                 POS_SYS_UNLOCK_DEVICE, 0, 0) == -1)
                      {
                        fprintf(stderr, "Error unlocking %s: errno = %s\n",
                                devInfo[MSR].cls_name,
                                GetTableEntry(ErrnoTable, PosErrno()));
                      }
                  }
              }

            break;
          }

        /**********************************************************/
        /*                      Keyboard                          */
        /**********************************************************/
        case PosTYPE_KEYBOARD:
          {
            if (DevNumber == PosDEVICE_POS_KEYBOARD_A ||
                DevNumber == PosDEVICE_CHECKOUT_KEYBOARD_A ||
                DevNumber == PosDEVICE_ANPOS_KEYBOARD_A)
              {
                if (devInfo[KBD].devdes == -1) 
                  {
                    rc = OpenAndAcquire(devInfo[KBD].devdes,
                                        devInfo[KBD].dev_name,
                                        devInfo[KBD].cls_name,
                                        openargs,
                                        PosNumber(openargs));
       
                    if (rc > 0)
                      {
                        devInfo[KBD].devdes = rc;
                        devInfo[KBD].dev_type = DevType;
                        devInfo[KBD].dev_subtype = SubType;
                        devInfo[KBD].dev_number = DevNumber;
       
                        if (PosIOCtl(devInfo[KBD].devdes,
                                     POS_SYS_UNLOCK_DEVICE, 0, 0) == -1)
                          {
                            fprintf(stderr, "Error unlocking %s: errno = %s\n",
                                            devInfo[KBD].cls_name,
                                            GetTableEntry(ErrnoTable, PosErrno()));
                          }
                        else
                          {
                            switch (DevNumber)
                              {
                                case PosDEVICE_ANPOS_KEYBOARD_A:
                                  {
                                    sioKeys = anposKeys;
                                    break;
                                  }
                                case PosDEVICE_CHECKOUT_KEYBOARD_A:
                                  {
                                    sioKeys = checkoutKeys;
       
                                    /*
                                     * Change the keyboard layout for the
                                     * original chekout keyboard to that
                                     * which was shipped from the factory.
                                     */
                                    SetCheckoutNumpadStyle(devInfo[KBD].devdes);
                                    break;
                                  }
                                case PosDEVICE_POS_KEYBOARD_A:
                                  {
                                    switch (SubType)
                                      {
                                        case PosKBD_SUBTYPE_ANPOS_2:
                                          {
                                            sioKeys = anposKeys;
                                            break;
                                          }
                                        case PosKBD_SUBTYPE_CHECKOUT_2:
                                          {
                                            sioKeys = checkout2Keys;
                                            break;
                                          }
                                        case PosKBD_SUBTYPE_MODIFIABLE_LAYOUT:
                                          {
                                            sioKeys = modifiableKeys;
                                            break;
                                          }
                                        default:
                                          {
                                            sioKeys = anposKeys;
                                            break;
                                          }
                                      }
                                    break;
                                  }
                                default:
                                  {
                                    sioKeys = anposKeys;
                                    break;
                                  }
                              }
                          }
                      }
                  }
              }

            break;
          }

        /**********************************************************/
        /*                   Cash Drawer (Till)                   */
        /**********************************************************/
        case PosTYPE_CASH_DRAWER:
          {
            if (devInfo[TIL].devdes == -1) 
              {
                rc = OpenAndAcquire(devInfo[TIL].devdes,
                                    devInfo[TIL].dev_name,
                                    devInfo[TIL].cls_name,
                                    openargs,
                                    PosNumber(openargs));
       
                if (rc > 0)
                  {
                    devInfo[TIL].devdes = rc;
                    devInfo[TIL].dev_type = DevType;
                    devInfo[TIL].dev_subtype = SubType;
                    devInfo[TIL].dev_number = DevNumber;
                  }
              }

            break;
          }

        /**********************************************************/
        /*                         NVRAM                          */
        /**********************************************************/
        case PosTYPE_NVRAM:
          {
            if (DevNumber == PosDEVICE_NVRAM_LOCAL)
              {
                if (devInfo[RAM].devdes != -1) 
                  {
                    rc = OpenAndAcquire(devInfo[RAM].devdes,
                                        devInfo[RAM].dev_name,
                                        devInfo[RAM].cls_name,
                                        openargs,
                                        PosNumber(openargs));
       
                    if (rc > 0)
                      {
                        devInfo[RAM].devdes = rc;
                        devInfo[RAM].dev_type = DevType;
                        devInfo[RAM].dev_subtype = SubType;
                        devInfo[RAM].dev_number = DevNumber;
       
                        /* Initialize the NVRAM */
                        ZeroNVRAM();
                      }
                  }
              }

            break;
         }

#if defined(POSTOUCH)
        /**********************************************************/
        /*                    TOUCH SCREEN                        */
        /**********************************************************/
       case PosTYPE_TOUCH_SCREEN:
         {
            if (devInfo[TCH].devdes == -1) 
              {
                rc = OpenAndAcquire(devInfo[TCH].devdes,
                                    devInfo[TCH].dev_name,
                                    devInfo[TCH].cls_name,
                                    openargs,
                                    PosNumber(openargs));
       
                if (rc > 0)
                  {
                    devInfo[TCH].devdes = rc;
                    devInfo[TCH].dev_type = DevType;
                    devInfo[TCH].dev_subtype = SubType;
                    devInfo[TCH].dev_number = DevNumber;
       
                    /*
                     * Initialize the touch screen user interface.
                     */
                    if (!InitTouchScreen(SubType)) /* 1st time online? */
                      {
                        FILE *pfl;
       
                        /*
                         * Load text into the 'ABOUT' dialog.
                         */
                        if (NULL != (pfl = fopen("checkout.abt","r")))
                          {
                            char buf[81] = {0};
                            int c;
       
                            while (fgets(buf,sizeof(buf),pfl))
                              {
                                /*
                                 *  If the line was longer than the buffer
                                 *  then dispose of the excess.
                                 */
                                if (!strchr(buf,'\n'))
                                  {
                                    while ((c = getc(pfl)) != '\n' && c != EOF)
                                      {
                                        /* null body */
                                      }
                                  }
       
                                WriteAbout(buf);
                              }
       
                            fclose(pfl);
                          }
       
                        /*
                         * Write greeting to touch screen's 2x20 mirror.
                         */
                        WriteTouchMirror(0, "                                        ", 40);
                        WriteTouchMirror(0, achGreeting, strlen(achGreeting));
                      }
                  }
              }

            break;
         }
#endif

       default:
         {
           break;
         }
      }

    return;
}


/********************************************************************/
/* TooManyItems                                                     */
/********************************************************************/

void TooManyItems(keyboard *kbd)
{
    DisplayString(devInfo[DSP].devdes, "Too many items", 0);
    DisplayString(devInfo[DSP].devdes, "Please press TOTAL", 20);

    /* pos_state = Tender_State; */

    /*
     * clear keyboard buffer
     */
    memset(kbd->buffer, '\0', kbd->index);
    kbd->index = 0;

    return;
}


/********************************************************************/
/* SignOff                                                          */
/********************************************************************/

void SignOff(transactionRecord *tran)
{
    #if (AUTO_SIGNON == 1)
      if (pos_state == Signed_Off_State)
        {
          UnlockDevice(devInfo[SCN].devdes);
        }
    #endif

    if (devInfo[PRT].devdes != -1)
      {
        /* Print Date and Time */
        WritePrinterDate(tran->transact_num);

        /* Chase Print on CR */
        WriteToCRstation(chase_print);

        pos_state = Till_Closed_State;
      }
    else
      {
        WriteTransactionLog(tran);

        pos_state = Signed_Off_State;

      }

    return;
}


/********************************************************************/
/* SignOn                                                           */
/********************************************************************/

int SignOn(transactionRecord *tran, keyboard *kbd)
{
    int rc = 0;

    time_t ltime = time((time_t *) 0);

    /*
     * Are we in the correct state to process this request?
     */
    if (pos_state == Signed_Off_State)
      {
        /*
         * Write the prompt to the display
         */

         /* DisplayFullScreen(devInfo[DSP].devdes, "Enter Item", 0); */

        /*
         * Print header on CR station and touch screen's
         * virtual CR station.
         */
        WriteToCRstation(" \x1b\x17       IBM\r\n\x1b\x3b");
        WriteToCRstation("\x1b\x45Point of Sale Subsystem\x1b\x46\r\n\n");

#if defined(POSTOUCH)
        WriteTouchStation(NULL);
        WriteTouchStation("           I B M");
        WriteTouchStation("  Point of Sale Subsystem");
        WriteTouchStation(" ");
#endif

        if (devInfo[PRT].dev_number != PosDEVICE_PRINTER_4610)
          {
            /*
             * Print sign-on on SJ station
             */
            WriteToSJstation("Operator Sign-On\r\n");
          }

        tran->transact_num++;

        /*
         * Print sign-on to transaction log
         */
        sprintf(tran->signOnBuf, "SIGN-ON %05d %.24s",
                                 tran->transact_num, ctime(&ltime));

        /*
         * Remove the new-line character
         */
        tran->signOnBuf[strlen(tran->signOnBuf)] = '\0';

        /*
         * Initialize the members of the transaction structure.
         */
        tran->item_list[0]       = '\0';
        tran->log_count          = 0;
        tran->log_tax_amount     = 0;
        tran->log_cash_amount    = 0;
        tran->log_credit_amount  = 0;
        total_amount             = 0;
        cash_tend_switch         = FALSE;

        /*
         * clear keyboard buffer
         */
        memset(kbd->buffer, '\0', sizeof(kbd->buffer));
        kbd->index = 0;

        /*
         * Unlock the scanner
         */
        if (devInfo[SCN].devdes != -1) 
          {
            UnlockDevice(devInfo[SCN].devdes);
          }

        welcome_displayed = FALSE;
        pos_state = Transaction_State;
      }
    else
      {
        /*
         * No, we are not in the correct state.  Therefore, sound
         * the keyboard tone and ignore this request.
         */
        rc = -1;
        SoundTone();
      }

    return rc;
}


/********************************************************************/
/* EnterKey                                                         */
/********************************************************************/

int EnterKey(transactionRecord *tran, keyboard *kbd)
{
    int rc = 0;
    char item_buff[21];
    char item_name[21];
    char buffer[256];
    long item_amount;

    switch (pos_state)
      {
        case Keyboard_Item_State:
          {
            if (tran->log_count >= MaxItems)
              {
                TooManyItems(kbd);
              }
            else
              {
                /*
                 * Get the information for the item entered
                 */
                item_amount = ItemLookup(kbd->buffer, item_name);

                /*
                 * Write the item information to the display
                 */
                DisplayFullScreen(devInfo[DSP].devdes, item_name, 0);
                DisplayAmount(devInfo[DSP].devdes, item_amount);

                /*
                 * Add the item amount to the total transaction amount
                 */
                total_amount += item_amount;

                /*
                 * Print item name and price on CR station
                 */
                if (strcmp("ITEM NOT FOUND",item_name))
                  {
                    /*
                     * Format the data to be printed
                     */
                    sprintf(buffer, "%-25s%10.2f\r\n", item_name,
                                   (double) item_amount / (double) 100);

                    if ((sjPrintItems == TRUE) &&
                        (devInfo[PRT].dev_number != PosDEVICE_PRINTER_4610))
                      {
                        WriteToCRandSJstations(buffer);
                      }
                    else
                      {
                        WriteToCRstation(buffer);
                      }

#if defined(POSTOUCH)
                    /*
                     * Format and print the data on the
                     * touch screen's virtual CR station.
                     */
                    sprintf(buffer,
                            " %-20s  %7.2f", item_name,
                            (double)item_amount / (double)100);
                    WriteTouchStation(buffer);
#endif

                    /*
                     * Add the item amount to the transaction logs item
                     * list.
                     */
                    sprintf(item_buff, "%8ld", item_amount);
                    strcat(tran->item_list, item_buff);

                    /*
                     * Increment the number of items we have
                     */
                    tran->log_count++;
                  }

                /*
                 * clear keyboard buffer
                 */
                memset(kbd->buffer, '\0', kbd->index);
                kbd->index = 0;

                /*
                 * Unlock the scanner
                 */
                UnlockDevice(devInfo[SCN].devdes);

                pos_state = Transaction_State;
              }

            break;
          }

        case Keyboard_Tender_State:
          {
            long tend_amount = atol(kbd->buffer);

            /*
             * If more cash has been received than was
             * required, determine how much should be returned
             */
            if (tend_amount >= total_amount)
              {
                long change_amount = tend_amount - total_amount;

                /*
                 * Write to display
                 */
                DisplayFullScreen(devInfo[DSP].devdes, "Change Due", 0);
                DisplayAmount(devInfo[DSP].devdes, change_amount);

                /*
                 * Print cash tend and change due on CR and SJ
                 */
                WriteToCRstation("\r\n");

                /*
                 * Format the data to be printed
                 */
                sprintf(buffer, "%-25s%10.2f\r\n%-25s%10.2f\r\n",
                          "Cash Tendered", (double) tend_amount / (double) 100,
                          "Change Due", (double) change_amount / (double) 100);

                if (devInfo[PRT].dev_number != PosDEVICE_PRINTER_4610)
                  {
                    WriteToCRandSJstations(buffer);
                  }
                else
                  {
                    WriteToCRstation(buffer);
                  }


#if defined(POSTOUCH)
                /*
                 * Format and write the data to the
                 * touch screen's virtual CR station.
                 */
                sprintf(buffer,
                        " %-20s  %7.2f",
                        "Cash Tendered",
                        (double)tend_amount / (double)100);
                WriteTouchStation(buffer);

                sprintf(buffer,
                        " %-20s  %7.2f",
                        "Change Due",
                        (double)change_amount / (double)100);
                WriteTouchStation(buffer);
#endif


                if (devInfo[TIL].devdes > 0)
                  {
                    pos_state = Cash_Tendered_State;

                    /*
                     * Open the cash drawer
                     */
                    OpenCashDrawer(devInfo[TIL].devdes);
                  }
                else
                  {
                    SignOff(tran);
                  }

                /*
                 * Update transaction log
                 */
                tran->log_cash_amount += total_amount;

                /*
                 * Update NVRAM
                 */
                UpdateNVRAM(devInfo[RAM].devdes,
                            total_amount, (long) 0, (long) 0);
              }
            else
              {
                /*
                 * Determine how much more needs to be received
                 * to satisify this transaction.
                 */
                total_amount -= tend_amount;

                /*
                 * Write to the display
                 */
                DisplayFullScreen(devInfo[DSP].devdes, "Balance Due", 0);
                DisplayAmount(devInfo[DSP].devdes, total_amount);

                /*
                 * Print cash tendered and total on CR and SJ
                 */
                WriteToCRstation("\r\n");

                /*
                 * Format the data to be printed
                 */
                sprintf(buffer, "%-25s%10.2f\r\n%-25s%10.2f\r\n",
                          "Cash Tendered", (double) tend_amount / (double) 100,
                          "Balance Due", (double) total_amount / (double) 100);
                if (devInfo[PRT].dev_number != PosDEVICE_PRINTER_4610)
                  {
                    WriteToCRandSJstations(buffer);
                  }
                else
                  {
                    WriteToCRstation(buffer);
                  }

#if defined(POSTOUCH)
                /*
                 * Format and write the data to the
                 * touch screen's virtual CR station.
                 */
                sprintf(buffer,
                        " %-20s  %7.2f",
                        "Cash Tendered",
                        (double)tend_amount / (double)100);
                WriteTouchStation(buffer);

                sprintf(buffer,
                        " %-20s  %7.2f",
                        "Balance Due",
                        (double)total_amount / (double)100);
                WriteTouchStation(buffer);
#endif


                /*
                 * clear keyboard buffer
                 */
                memset(kbd->buffer, '\0', kbd->index);
                kbd->index = 0;

                /*
                 * Update NVRAM
                 */
                UpdateNVRAM(devInfo[RAM].devdes,
                            tend_amount, (long) 0, (long) 0);

                pos_state              = Tender_State;
                tran->log_cash_amount += tend_amount;
                cash_tend_switch       = TRUE;
              }

            break;
          }

        case Printer_Error_State:
          {
            /*
             * If the printer is not opened and acquired, ignore
             * this request.
             */
            if (devInfo[PRT].devdes != -1)
              {
                /*
                 * Try the print command again
                 */
                if (PosIOCtl(devInfo[PRT].devdes, POS_PRN_RETRY_PRINTING,
                             NULL, 0) == -1)
                  {
                    fprintf(stderr, "PosIOCtl (RETRY) Printer: errno = %s\n",
                                    GetTableEntry(ErrnoTable, PosErrno()));
                  }
              }

            /*
             * Remove the error message from the display
             */
            ClearDisplay(devInfo[DSP].devdes);

            pos_state = sav_state;
            break;
          }

        default:
          {
            /*
             * Invalid state for processing, ignore this request
             */
            rc = -1;
            SoundTone();
            break;
          }
      }

    return rc;
}


/********************************************************************/
/* ClearKey                                                         */
/********************************************************************/

int ClearKey(keyboard *kbd)
{
    int rc = 0;

    switch (pos_state)
      {
        case Transaction_State:
        case Keyboard_Item_State:
          {
            /*
             * Clear the keyboard buffer
             */
            memset(kbd->buffer, '\0', kbd->index);
            kbd->index = 0;

            /*
             * Write to the display
             */
            DisplayFullScreen(devInfo[DSP].devdes, "Enter Item", 0);
            DisplayItemRight(devInfo[DSP].devdes, kbd->buffer);

            pos_state = Transaction_State;
            break;
          }

        case Keyboard_Tender_State:
          {
            /*
             * Clear the keyboard buffer
             */
            memset(kbd->buffer, '\0', kbd->index);
            kbd->index = 0;

            /*
             * Write to the display
             */
            DisplayFullScreen(devInfo[DSP].devdes, "Enter Amount", 0);
            DisplayAmount(devInfo[DSP].devdes, 0);

            break;
          }

        default:
          {
            /*
             * Invalid state for processing, ignore this request
             */
            rc = -1;
            SoundTone();
            break;
          }
      }

    return rc;
}


/********************************************************************/
/* TotalKey                                                         */
/********************************************************************/

int TotalKey(transactionRecord *tran)
{
    int rc = 0;

    char buffer[256];
    long sub_total = total_amount;

    if (pos_state == Transaction_State || tran->log_count >= MaxItems)
      {
        long tax_amount = (total_amount * Tax_Percent) / 100;

        total_amount += tax_amount;

        if (total_amount != 0)
          {
            /*
             * Write to the display
             */
            DisplayFullScreen(devInfo[DSP].devdes, "Total", 0);
            DisplayAmount(devInfo[DSP].devdes, total_amount);

            /*
             * Print tax and total on CR and SJ stations
             */
            WriteToCRstation("\r\n");

            /*
             * Format the data to be printed
             */
            sprintf(buffer,
                "%-25s%10.2f\r\n%-25s%10.2f\r\n\x1b\x45%-25s%10.2f\x1b\x46\r\n",
                "Sub-Total", (double) sub_total    / (double) 100,
                "Tax",       (double) tax_amount   / (double) 100,
                "Total",     (double) total_amount / (double) 100);
            if (devInfo[PRT].dev_number != PosDEVICE_PRINTER_4610)
              {
                WriteToCRandSJstations(buffer);
              }
            else
              {
                WriteToCRstation(buffer);
              }


#if defined(POSTOUCH)
            /*
             * Format and write the data to the
             * touch screen's virtual CR station.
             */
            WriteTouchStation(" ");

            sprintf(buffer,
                    " %-20s  %7.2f",
                    "Sub-Total",
                    (double)sub_total / (double)100);
            WriteTouchStation(buffer);

            sprintf(buffer,
                    " %-20s  %7.2f",
                    "Tax",
                    (double)tax_amount / (double)100);
            WriteTouchStation(buffer);

            sprintf(buffer,
                    " %-20s  %7.2f",
                    "Total",
                    (double)total_amount / (double)100);
            WriteTouchStation(buffer);

            WriteTouchStation(" ");
#endif

            /*
             * Update NVRAM
             */
            UpdateNVRAM(devInfo[RAM].devdes, (long) 0, (long) 0, tax_amount);

            /*
             * Update transaction log info
             */
            tran->log_tax_amount = tax_amount;

            pos_state = Tender_State;
          }
        else
          {
            /*
             * Write to display
             */
            DisplayFullScreen(devInfo[DSP].devdes, "No Sale", 6);

            /*
             * Write the "No Sale" message to the CR and SJ Stations
             */
            if (devInfo[PRT].dev_number != PosDEVICE_PRINTER_4610)
              {
                WriteToCRandSJstations("\r\n     NO SALE\r\n");
              }
            else
              {
                WriteToCRstation("\r\n     NO SALE\r\n");
              }


#if defined(POSTOUCH)
            WriteTouchStation("    NO SALE");
#endif

            /*
             * Print date and time on CR and SJ
             */
            WritePrinterDate(tran->transact_num);

            if (devInfo[PRT].devdes != -1)
              {
                /*
                 * Chase Print on CR
                 */
                WriteToCRstation(chase_print);

                pos_state = Till_Closed_State;
              }
            else
              {
                pos_state = Signed_Off_State;
              }
          }

        #if (AUTO_SIGNON != 1)
          /* Lock the HHBCR */
          LockDevice(devInfo[SCN].devdes);
        #endif

        tran->log_count = 0;
      }
    else
      {
        /*
         * Invalid state for processing, ignore this request
         */
        rc = -1;
        SoundTone();
      }

    return rc;
}


/********************************************************************/
/* CashKey                                                          */
/********************************************************************/

int CashKey(transactionRecord *tran)
{
    int rc = 0;

    if (pos_state == Tender_State && tran->log_count == 0)
      {
        /*
         * Write to the display
         */
        DisplayFullScreen(devInfo[DSP].devdes, "Enter Amount", 0);
        DisplayAmount(devInfo[DSP].devdes, total_amount);

        pos_state = Keyboard_Tender_State;
      }
    else
      {
        /*
         * Invalid state for processing, ignore this request
         */
        rc = -1;
        SoundTone();
      }

    return rc;
}


/********************************************************************/
/* ChargeKey                                                        */
/********************************************************************/

int ChargeKey(transactionRecord *tran)
{
    int rc = 0;

    if (pos_state           == Tender_State &&
        tran->log_count     == 0 &&
        devInfo[MSR].devdes >  0)
      {
        /*
         * Write to the display
         */
        DisplayFullScreen(devInfo[DSP].devdes, "Slide Card", 0);

        /*
         * Unlock the MSR
         */
        UnlockDevice(devInfo[MSR].devdes);

        pos_state = MSR_Tender_State;
      }
    else
      {
        /*
         * Invalid state for processing, ignore this request
         */
        rc = -1;
        SoundTone();
      }

    return rc;
}


/********************************************************************/
/* CheckKey                                                         */
/********************************************************************/

int CheckKey(transactionRecord *tran)
{
    int rc = 0;

    if (pos_state           == Tender_State &&
        tran->log_count     == 0 &&
        devInfo[PRT].devdes != -1 &&
        MicrInstalled == TRUE)
      {
        /*
         * Write to the display
         */
        DisplayFullScreen(devInfo[DSP].devdes, "Insert Check", 0);

        pos_state = Check_Tender_State;
      }
    else
      {
        /*
         * Invalid state for processing, ignore this request
         */
        rc = -1;
        SoundTone();
      }

    return rc;
}


/********************************************************************/
/* EndProgramKey                                                    */
/********************************************************************/

int EndProgramKey(void)
{
    int rc = 0;

    long close_cash;
    long close_credit;
    long close_tax;

    /*
     * Write the closed message to the display
     */
    DisplayFullScreen(devInfo[DSP].devdes, "    ** Closed **", 0);

    /*
     * If the printer has been opened and acquired, print the total
     * information to the SJ station.
     */
    if (devInfo[PRT].devdes > 0)
      {
        /*
         * Get the totals stored in NVRAM
         */
        ReadNVRAM(&close_cash, &close_credit, &close_tax);

        /*
         * Write the cash sales, credit sales, and total tax amount
         * to the SJ
         */
        WritePrinterClose("Cash Sales", close_cash);
        WritePrinterClose("Credit Sales", close_credit);
        WritePrinterClose("Tax Collected", close_tax);

        /*
         * Chase Print on SJ
         */
        if (devInfo[PRT].dev_number != PosDEVICE_PRINTER_4610)
          {
            WriteToSJstation(chase_print);
          }
        else
          {
            WriteToCRstation(chase_print);
          }
      }
//    else
//      {
        exit_switch = TRUE;
//      }

    pos_state = Store_Closing_State;

    return rc;
}


/********************************************************************/
/* ProcessCharacter                                                 */
/********************************************************************/

void ProcessCharacter(POSQMSG *qmsg, transactionRecord *tran, keyboard *kbd)
{
    unsigned char keyFlags = (unsigned short)((qmsg->mp2 >> 16) & 0xFFFF);

    /*
     * Ignore key up messages
     */
    if (keyFlags & PosKC_KEYUP)
      {
        /*
         * Ignore the character that comes in when the
         * key is lifted.  We have already processed
         * the key when it was pressed.
         */
      }
    else
      {
        unsigned char scanCode = ((unsigned char) (qmsg->mp2 & 0xFF));
        unsigned char keyCode  = sioKeys[scanCode];
        ProcessKeyCode(keyCode, tran, kbd);
      }
}


/********************************************************************/
/* ProcessKeyCode                                                   */
/********************************************************************/

int ProcessKeyCode(unsigned char keyCode, transactionRecord *tran, keyboard *kbd)
{
    int rc = 0;

    #if (AUTO_SIGNON == 1)
      if (pos_state == Signed_Off_State)
        {
          SignOn(tran, kbd);
        }
    #endif

    switch (pos_state)
      {
        case Tender_State:
          {
            switch (keyCode)
              {
                case CASH:
                  {
                    rc = CashKey(tran);
                    break;
                  }

                case CHARGE:
                  {
                    rc = ChargeKey(tran);
                    break;
                  }

                case CHECK:
                  {
                    rc = CheckKey(tran);
                    break;
                  }

                default:
                  {
                    rc = -1;
                    SoundTone();
                    DisplayFullScreen(devInfo[DSP].devdes,
                                      "Select Tender Type",
                                      0);
                    break;
                  }
              }

            break;
          }

        default:
          {
            switch (keyCode)
              {
                case NUM0:
                case NUM1:
                case NUM2:
                case NUM3:
                case NUM4:
                case NUM5:
                case NUM6:
                case NUM7:
                case NUM8:
                case NUM9:
                  {
                    rc = ProcessNumber(tran, kbd, keyCode);
                    break;
                  }

                case SIGN_ON:
                  {
                    rc = SignOn(tran, kbd);
                    break;
                  }

                case ENTER:
                  {
                    rc = EnterKey(tran, kbd);
                    break;
                  }

                case CLEAR:
                  {
                    rc = ClearKey(kbd);
                    break;
                  }

                case TOTAL:
                  {
                    rc = TotalKey(tran);
                    break;
                  }

                case CASH:
                  {
                    rc = CashKey(tran);
                    break;
                  }

                case CHARGE:
                  {
                    rc = ChargeKey(tran);
                    break;
                  }

                case CHECK:
                  {
                    rc = CheckKey(tran);
                    break;
                  }

                case PRINT:
                  {
                    rc = WriteFeatures();
                    break;
                  }

                case END_PRG:
                  {
                    rc = EndProgramKey();
                    break;
                  }

                default:
                  {
                    /*
                     * Invalid state for processing, ignore this request
                     */
                    rc = -1;
                    SoundTone();
                    break;
                  }
              }
          }
      }

    return rc;
}


/********************************************************************/
/* ProcessNumber                                                    */
/********************************************************************/

int ProcessNumber(transactionRecord *tran, keyboard *kbd,
                  unsigned char keyCode)
{
    int rc = 0;

    switch (pos_state)
      {
        case Keyboard_Item_State:
          {
            if (kbd->index < 18)
              {
                /*
                 * Save the character
                 */
                kbd->buffer[kbd->index++] = keyCode;

                /*
                 * Write to the display
                 */
                DisplayItemRight(devInfo[DSP].devdes, kbd->buffer);
              }
            else
              {
                /*
                 * Invalid state for processing, ignore this request
                 */
                rc = -1;
                SoundTone();
              }
            break;
          }

        case Transaction_State:
          {
            /*
             * Clear the display
             */
            ClearDisplay(devInfo[DSP].devdes);

            /*
             * Save the character
             */
            kbd->buffer[kbd->index++] = keyCode;

            /*
             * Write to the display
             */
            DisplayItemRight(devInfo[DSP].devdes, kbd->buffer);

            /*
             * Lock the scanner
             */
            if (devInfo[SCN].devdes != -1)
              {
                LockDevice(devInfo[SCN].devdes);
              }

            pos_state = Keyboard_Item_State;

            break;
          }

        case Keyboard_Tender_State:
          {
            if (kbd->index < 9)
              {
                /*
                 * Save the character
                 */
                kbd->buffer[kbd->index++] = keyCode;

                /*
                 * Write to the display
                 */
                DisplayAmount(devInfo[DSP].devdes, atol(kbd->buffer));
              }
            else
              {
                /*
                 * Invalid state for processing, ignore this request
                 */
                rc = -1;
                SoundTone();
              }
            break;
          }

        default:
          {
            /*
             * Invalid state for processing, ignore this request
             */
            rc = -1;
            SoundTone();
            break;
          }
      }

    return rc;
}


/********************************************************************/
/* ProcessScanData                                                  */
/********************************************************************/

void ProcessScanData(int devdes, transactionRecord *tran, keyboard *kbd)
{
    char buffer[60];
    int  scnDataLen = SCN_Buf_Size;
    union
      {
        PosScannerDataHdr scnHdr;
        char scnBuffer[SCN_Buf_Size];
      } scnData;

    #if (AUTO_SIGNON == 1)
      if (pos_state == Signed_Off_State)
        {
          SignOn(tran, kbd);
        }
    #endif

    if (pos_state == Transaction_State)
      {
        /*
         * Lock the scanner so that we don't get unexpected data
         */
        LockDevice(devdes);

        /*
         * If we have reached the maximum number of items for this
         * transaction, display the message and force the operator
         * to go to the total state.
         */
        if (tran->log_count >= MaxItems)
          {
            TooManyItems(kbd);
          }
        else
          {
            /*
             * Read from the scanner
             */
            if (ReadScanner(devdes, scnData.scnBuffer, &scnDataLen))
              {
                long item_amount;
                char item_name[21];

                scnData.scnBuffer[scnData.scnHdr.LblLen + 8] = '\0';
                item_amount = ItemLookup(&scnData.scnBuffer[8], item_name);

                /*
                 * Write to the display
                 */
                DisplayFullScreen(devInfo[DSP].devdes, item_name, 0);
                DisplayAmount(devInfo[DSP].devdes, item_amount);

                if (strcmp("ITEM NOT FOUND",item_name))
                  {
                    char item_buff[21];

                    /*
                     * Print item name and price on CR
                     */
                    sprintf(buffer, "%-25s%10.2f\r\n", item_name,
                                    (double) item_amount / (double) 100);
                    if (devInfo[PRT].dev_number != PosDEVICE_PRINTER_4610) 
                      {
                        WriteToCRandSJstations(buffer);
                      }
                    else
                      {
                        WriteToCRstation(buffer);
                      }

                    sprintf(item_buff, "%8ld", item_amount);
                    strcat(tran->item_list, item_buff);

#if defined(POSTOUCH)
                    /*
                     * Format and write the data to the
                     * touch screen's virtual CR station.
                     */
                    sprintf(buffer,
                            " %-20s%  7.2f",
                            item_name,
                            (double)item_amount / (double)100);
                    WriteTouchStation(buffer);
#endif

                    /*
                     * Update our transaction count
                     */
                    tran->log_count++;
                  }

                total_amount += item_amount;

                /*
                 * clear keyboard buffer
                 */
                memset(kbd->buffer, '\0', kbd->index);
                kbd->index = 0;

                /*
                 * Unlock the scanner
                 */
                UnlockDevice(devdes);
              }
          }
      }
    else
      {
        SoundTone();

        if (devdes > 0)
          {
            /*
             * Flush the scanner data
             */
            if (PosIOCtl(devdes, POS_SCAN_DISCARD_DATA, NULL, 0) == -1)
              {
                fprintf(stderr, "PosIOCtl (DISCARD_DATA) Scanner: errno = %s\n",
                                GetTableEntry(ErrnoTable, PosErrno()));
              }
          }
      }

    return;
}


#if defined(POSTOUCH)
/********************************************************************/
/* ProcessTouchData                                                 */
/********************************************************************/
void ProcessTouchData(POSQMSG *qmsg, transactionRecord *tran, keyboard *kbd)
{
  /*
   * Handle device error.
   */
  if (qmsg->mp1 & (PosTOUCH_ERROR << 16))
    {
      /* Add error handler here. */
    }

  /*
   * Process touch going active.
   */
  else if (qmsg->mp1 & (PosTOUCH_BUTTON1_DOWN << 16))
    {
      ProcessTouchActive(qmsg,tran,kbd);
    }

  /*
   * Process touch going inactive.
   */
  else if (!(qmsg->mp1 & (PosTOUCH_BUTTON1_DOWN << 16)))
    {
      /* Ignore. */
    }
}
#endif


/********************************************************************/
/* ProcessMsrData                                                   */
/********************************************************************/

void ProcessMsrData(int devdes, transactionRecord *tran)
{
    char buffer[60];
    int  msrDataLen = MSR_Buf_Size;
    union
      {
        PosMsrDataHdr msrHdr;
        char msrBuffer[MSR_Buf_Size];
      } msrData;

    if (pos_state == MSR_Tender_State)
      {
        /*
         * Clear the information from the display
         */
        ClearDisplay(devInfo[DSP].devdes);

        /*
         * Print credit amount on CR and SJ stations
         */
        sprintf(buffer, "\r\n%-25s%10.2f\r\n", "Credit Amount",
                        (double) total_amount / (double) 100);
        if (devInfo[PRT].dev_number != PosDEVICE_PRINTER_4610) 
          {
            WriteToCRandSJstations(buffer);
          }
        else
          {
            WriteToCRstation(buffer);
          }

#if defined(POSTOUCH)
        /*
         * Format and write the data to the
         * touch screen's virtual CR station.
         */
        sprintf(buffer,
                " %-20s  %7.2f",
                "Credit Amount",
                (double)total_amount / (double)100);
        WriteTouchStation(buffer);
#endif

        /*
         * Read the MSR data
         */
        ReadMSR(devdes, msrData.msrBuffer, &msrDataLen);

        while ((msrDataLen > 0) && (msrData.msrHdr.Track != PosTRACK_2))
          {
            msrDataLen -= (msrData.msrHdr.Length + sizeof(msrData.msrHdr));
            memcpy(msrData.msrBuffer,
                   &(msrData.msrBuffer[(msrData.msrHdr.Length +
                                        sizeof(msrData.msrHdr))]),
                   msrDataLen);
          }

        if (msrDataLen > sizeof(msrData.msrHdr))
          {
            int index;

            for (index = sizeof(msrData.msrHdr); index < msrDataLen; index++)
              {
                if (msrData.msrBuffer[index] > 9)
                  {
                    msrData.msrBuffer[index] = '.';
                  }
                else
                  {
                    msrData.msrBuffer[index] += '0';
                  }
              }

            msrData.msrBuffer[index] = '\0';
          }

        if (strlen(&(msrData.msrBuffer[sizeof(msrData.msrHdr)]))
                     > Max_Print_Width)
          {
            msrData.msrBuffer[Max_Print_Width] = '\0';
          }

        /*
         * Print account number on CR and SJ stations
         */
        if (devInfo[PRT].dev_number != PosDEVICE_PRINTER_4610)
          {
            WriteToCRandSJstations(&(msrData.msrBuffer[sizeof(msrData.msrHdr)]));
            WriteToCRandSJstations("\r\n");
          }
        else
          {
            WriteToCRstation(&(msrData.msrBuffer[sizeof(msrData.msrHdr)]));
            WriteToCRstation("\r\n");
          }

#if defined(POSTOUCH)
        WriteTouchStation(&(msrData.msrBuffer[sizeof(msrData.msrHdr)]));
#endif

        /*
         * Update transaction log information
         */
        tran->log_credit_amount = total_amount;

        if (!cash_tend_switch)
          {
            SignOff(tran);
          }
        else if (devInfo[TIL].devdes > 0)
          {
            /*
             * Open the cash drawer
             */
            OpenCashDrawer(devInfo[TIL].devdes);

            pos_state = Cash_Tendered_State;
          }
        else
          {
            /*
             * Write to the display
             */
            ClearDisplay(devInfo[DSP].devdes);

            SignOff(tran);
          }

        /*
         * Update NVRAM
         */
        UpdateNVRAM(devInfo[RAM].devdes, (long) 0, total_amount, (long) 0);
      }
    else
      {
        /*
         * Invalid state for processing, ignore this request
         */
        SoundTone();

        /*
         * Read the MSR data and discard it
         */
        ReadMSR(devdes, msrData.msrBuffer, &msrDataLen);
      }

    return;
}


/********************************************************************/
/* ProcessPrinterData                                               */
/********************************************************************/

void ProcessPrinterData(POSQMSG *qmsg,
                        int devdes,
                        transactionRecord *tran)
{
    long prnDataLen = (long) qmsg->mp2 & 0xFFFF;
    char buffer[256];

    if (pos_state == Check_Tender_State)
      {
        char displayBuffer[50];
        char relDocChar[] = { 0x1B, 0x4D };

        /*
         * Write to the display
         */
        DisplayFullScreen(devInfo[DSP].devdes, "Remove Check", 0);

         /*
          * Make sure we can write to the DI station
          */
        SetDIPrinting();
        WriteToDIstation(relDocChar, sizeof(relDocChar));

        /*
         * Read the printer information
         */
        ReadMicr(buffer, &prnDataLen);

        MicrInfoRequested = FALSE;

        /*
         * Get the check information ready for printing
         */
        memcpy(displayBuffer, buffer, prnDataLen);
        displayBuffer[prnDataLen] = '\0';
        strcat(displayBuffer, "\r\n");

        /*
         * Print check amount on CR, SJ, and touch stations.
         */
        sprintf(buffer, "\r\n%-25s%10.2f\r\n", "Check Amount",
                        (double) total_amount / (double) 100);
        if (devInfo[PRT].dev_number != PosDEVICE_PRINTER_4610)
          {
            WriteToCRandSJstations(buffer);
          }
        else
          {
            WriteToCRstation(buffer);
          }


#if defined(POSTOUCH)
        WriteTouchStation(buffer);
#endif

        /*
         * Print account number on CR, SJ, and touch stations.
         */
        if (devInfo[PRT].dev_number != PosDEVICE_PRINTER_4610)
          {
            WriteToCRandSJstations(displayBuffer);
          }
        else
          {
            WriteToCRstation(displayBuffer);
          }

#if defined(POSTOUCH)
        WriteTouchStation(displayBuffer);
#endif

        /*
         * Update transaction log information
         */
        tran->log_cash_amount = total_amount;

        /*
         * Print date and time on receipt and journal log
         */
        WritePrinterDate(tran->transact_num);

        /*
         * Chase print on CR station
         */
        WriteToCRstation(chase_print);

        /* Update NVRAM */
        UpdateNVRAM(devInfo[RAM].devdes, (long) 0, total_amount, (long) 0);

        pos_state = Check_Remove_State;
      }
    else
      {
        /*
         * Invalid state for processing, ignore this request
         */
        SoundTone();

        /*
         * Read the check information and discard it
         */
        ReadMicr(buffer, &prnDataLen);
      }

    return;
}


/********************************************************************/
/* ProcessPrinterStatusChange                                       */
/********************************************************************/

void ProcessPrinterStatusChange(POSQMSG *qmsg, int devdes)
{
    long change = (qmsg->mp1 >> 16) & 0xFFFF;
    long status = qmsg->mp2;

    if (MicrInstalled     == TRUE &&
        MicrInfoRequested == FALSE &&
        pos_state         == Check_Tender_State)
      {
        if (((change == PosCHANGE_DOCUMENT_AT_FRONT) &&
             (status  & PosSTATUS_DOCUMENT_AT_FRONT) &&
             !(status & PosSTATUS_DOCUMENT_READY)) ||
            ((change == PosCHANGE_DOCUMENT_AT_TOP) &&
             (status  & PosSTATUS_DOCUMENT_AT_TOP) &&
             !(status & PosSTATUS_DOCUMENT_READY)))
          {
            char regDocChar[] = { 0x1B, 0x4C };

            SetDIPrinting();

            WriteToDIstation(regDocChar, sizeof(regDocChar));
          }
        else if (change == PosCHANGE_DOCUMENT_REGISTER)
          {
            if (status & PosSTATUS_DOCUMENT_READY)
              {
                char micrReadChar[] = { 0x1B, 0x52 };

                SetDIPrinting();

                MicrInfoRequested = TRUE;
                WriteToDIstation(micrReadChar, sizeof(micrReadChar));

                /*
                 * Write to the display
                 */
                DisplayFullScreen(devInfo[DSP].devdes, "Reading Check", 0);
              }
          }
        else if ((change == PosCHANGE_DOCUMENT_AT_FRONT ||
                  change == PosCHANGE_DOCUMENT_AT_TOP   ||
                  change == PosCHANGE_DOCUMENT_REGISTER) &&
                 !(status & (PosSTATUS_DOCUMENT_READY |
                             PosSTATUS_DOCUMENT_AT_FRONT |
                             PosSTATUS_DOCUMENT_AT_TOP)))
          {
            SetCRPrinting();
            pos_state = Till_Closed_State;
          }
      }
    else if ((pos_state == Check_Remove_State) &&
             ((change == PosCHANGE_DOCUMENT_AT_FRONT ||
               change == PosCHANGE_DOCUMENT_AT_TOP   ||
               change == PosCHANGE_DOCUMENT_REGISTER) &&
              !(status & (PosSTATUS_DOCUMENT_READY |
                          PosSTATUS_DOCUMENT_AT_FRONT |
                          PosSTATUS_DOCUMENT_AT_TOP))))
      {
        SetCRPrinting();
        pos_state = Till_Closed_State;
      }

    return;
}


/********************************************************************/
/* ProcessChaseComplete                                             */
/********************************************************************/

void ProcessChaseComplete(POSQMSG *qmsg, transactionRecord *tran)
{
    int Station = (unsigned short)((unsigned long)qmsg->mp1 >> 16 & 0xFF);

    switch (pos_state)
      {
        case Till_Closed_State:
          {
            if (Station == PosCHASE_FROM_CR)
              {
                WriteTransactionLog(tran);

                pos_state = Signed_Off_State;
              }
            break;
          }

        case Store_Closing_State:
          {
            if (Station == PosCHASE_FROM_SJ)
              {
                if (tran->log_count != 0)
                  {
                    /* strcpy(signOffBuf, "*** Incomplete\n"); */
                    WriteTransactionLog(tran);
                  }

                exit_switch = TRUE;
              }
            break;
          }

        case Print_Features_State:
          {
            if (Station == PosCHASE_FROM_CR)
              {
                ClearDisplay(devInfo[DSP].devdes);
                welcome_displayed = FALSE;

                pos_state = Signed_Off_State;
              }
            break;
          }

        default:
          {
            break;
          }
      }

    return;
}


/********************************************************************/
/* ProcessPrinterError                                              */
/********************************************************************/

void ProcessPrinterError(POSQMSG *qmsg)
{
    char buffer[41];

    sprintf(buffer, "Printer Error       x%08x  x%08x", qmsg->mp1, qmsg->mp2);

    /*
     * Write to the display
     */
    DisplayFullScreen(devInfo[DSP].devdes, buffer, 0);

    sav_state = pos_state;
    pos_state = Printer_Error_State;
    MicrInfoRequested = FALSE;

    return;
}


/********************************************************************/
/* ProcessSku                                                       */
/********************************************************************/
int ProcessSku(char *szSku, transactionRecord *tran, keyboard *kbd)
{
  int rc = 0;
  long lItemAmount;
  char achItemName[21], achPrintString[60];

  #if (AUTO_SIGNON == 1)
    if (pos_state == Signed_Off_State)
      {
        SignOn(tran, kbd);
      }
  #endif


  switch (pos_state)
    {
      case Tender_State:
        {
          SoundTone();
          rc = -1;
          DisplayFullScreen(devInfo[DSP].devdes,
                            "Select Tender Type",
                            0);
          break;
        }

      case Transaction_State:
        {
          /*
           * If we have reached the maximum number of items for this
           * transaction, display the message and force the operator
           * to go to the total state.
           */
          if (tran->log_count >= MaxItems)
            {
              TooManyItems(kbd);
            }
          else
            {
              lItemAmount = ItemLookup(szSku, achItemName);

              /*
               * Write to the display
               */
              DisplayFullScreen(devInfo[DSP].devdes, achItemName, 0);
              DisplayAmount(devInfo[DSP].devdes, lItemAmount);

              if (strcmp("ITEM NOT FOUND",achItemName))
                {
                  char achAmountString[21];

                  /*
                   * Print item name and price on CR
                   */
                  sprintf(achPrintString, "%-25s%10.2f\r\n", achItemName,
                                  (double) lItemAmount / (double) 100);
                  if (devInfo[PRT].dev_number != PosDEVICE_PRINTER_4610) 
                    {
                      WriteToCRandSJstations(achPrintString);
                    }
                  else
                    {
                      WriteToCRstation(achPrintString);
                    }

#if defined(POSTOUCH)
                  /*
                   * Format and print the data on the
                   * touch screen's virtual CR station.
                   */
                  sprintf(achPrintString,
                          " %-20s  %7.2f",
                          achItemName,
                          (double)lItemAmount / (double)100);
                  WriteTouchStation(achPrintString);
#endif

                  sprintf(achAmountString, "%8ld", lItemAmount);

                  strcat(tran->item_list, achAmountString);

                  /*
                   * Update our transaction count
                   */
                  tran->log_count++;
                }

              total_amount += lItemAmount;
            }

          break;
        }

      default:
        {
          SoundTone();
          rc = -1;
          break;
        }
    }

  return rc;
}


/********************************************************************/
/* ProcessTillOpened                                                */
/********************************************************************/

void ProcessTillOpened(long *till_open_tm)
{
    if (pos_state == Cash_Tendered_State)
      {
        *till_open_tm = time((time_t *) 0) + 20;

        pos_state = Till_Opened_State;
      }

    return;
}


/********************************************************************/
/* ProcessTillClosed                                                */
/********************************************************************/

void ProcessTillClosed(transactionRecord *tran)
{
    if (pos_state == Till_Opened_State)
      {
        SignOff(tran);
      }

    return;
}


#if defined(__OS2__)
/********************************************************************/
/* SetProcessPriority                                               */
/********************************************************************/

void SetProcessPriority(int argc, char *argv[])
{
    unsigned   long priority_class = 0;
    long       priority_delta = 0;
    unsigned   long rc;

    if (argc >= 4)
      {
        priority_delta = atol(argv[3]);

        if (argc >= 5)
          {
            priority_class = atol(argv[4]);
          }

        if ((rc = DosSetPriority(PRTYS_PROCESS,
                                 priority_class,
                                 priority_delta,
                                 0)) != 0)
          {
            fprintf(stderr, "DosSetPriority: rc = %d\n", rc);
          }
      }

    return;
}
#else
void SetProcessPriority(int argc, char *argv[])
{
  /* don't know how to set priority in this environment */
}
#endif


/********************************************************************/
/* main                                                             */
/********************************************************************/

int main(int argc, char *argv[])
{
    ssize_t rc;
    int coerr;
    PosArg  TimeoutArg;
    long    till_open_tm = 0;
    transactionRecord tran;
    keyboard          kbd;
    char              *sjPrintEnv;
    FILE *pfl;

    memset(&tran, 0, sizeof(transactionRecord));
    memset(&kbd,  0, sizeof(keyboard));

    /*
     * Buffer stdout for curses' sake.
     * Immediately flush all stderr writes.
     */

    setvbuf(stdout, NULL, _IOFBF, 8192);
    setvbuf(stderr, NULL, _IONBF, 0);

    if (argc < 3)
      {
        fprintf(stderr, "Transaction Log Filename Missing\n");
        if (argc < 2)
          {
            fprintf(stderr, "Resource Filename Missing\n");
          }
        return 3;
      }

    printf("\n***** %s starting *****\n", argv[0]);
    fflush(stdout);

#if defined(__OS2__)
    SetProcessPriority(argc, argv);
#endif

    strcpy(tran.logFileName, argv[2]);

    TimeoutArg.name  = PosNreadTimeout;
    TimeoutArg.value = Sleep_ms;

    /* Initialize the subsystem. */
    if (PosInitialize(argv[0],       /* char *name   */
                      argv[1],       /* char *class  */
                      &argc,         /* int *argc    */
                      &argv[0],      /* char *argv[] */
                      &TimeoutArg,   /* ArgList args */
                      1) == -1)      /* int nargs    */
      {
        fprintf(stderr, "PosInitialize: errno = %s\n",
                        GetTableEntry(ErrnoTable, PosErrno()));
        return 4;
      }
    else
      {
        if ((item_rec_file = fopen("checkout.dat", "r")) == NULL)
          {
            fprintf(stderr, "fopen \"checkout.dat\" failed\n");

            return 3;
          }

        if ((sjPrintEnv = getenv("PRINT_ITEMS_ON_SJ")) != (char *) NULL)
          {
            if (strcmp(sjPrintEnv, "ON") == 0 ||
                strcmp(sjPrintEnv, "on") == 0)
              {
                sjPrintItems = TRUE;
              }
          }

        /*
         * Get greeting from file.
         */
        if (NULL != (pfl = fopen("checkout.grt","r")))
          {
            char *pch;

            fgets(achGreeting,sizeof(achGreeting),pfl);
            fclose(pfl);

            if (NULL != (pch = strchr(achGreeting,'\n')))
              {
                *pch = '\0';
              }
          }

        printf("***** initialize complete *****\n");
        fflush(stdout);
      }

    while (exit_switch == FALSE)
      {

        /* Check message queue for any input */
        while (((rc = PosRead(0, &qmsg, EVENT_MESSAGE_SIZE))
                                                      == EVENT_MESSAGE_SIZE) &&
               (exit_switch == FALSE))
          {
            switch (qmsg.msg)
              {
                case POSM_KBD_WM_CHAR:
                  {
                    ProcessCharacter(&qmsg, &tran, &kbd);
                    break;
                  }

                case POSM_SCAN_DATA_AVAIL:
                  {
                    int scandevdes = (int)(qmsg.mp1 & 0x000000FF);
                    ProcessScanData(scandevdes, &tran, &kbd);
                    break;
                  }

#if defined(POSTOUCH)
                case POSM_TOUCH_DATA:
                  {
                    ProcessTouchData(&qmsg, &tran, &kbd);
                    break;
                  }
#endif

                case POSM_MSR_DATA_AVAIL:
                  {
                    ProcessMsrData(devInfo[MSR].devdes, &tran);
                    break;
                  }

                case POSM_PRN_STATUS_CHANGE:
                  {
                    ProcessPrinterStatusChange(&qmsg, devInfo[PRT].devdes);
                    break;
                  }

                case POSM_PRN_DATA_AVAIL:
                  {
                    ProcessPrinterData(&qmsg, devInfo[PRT].devdes, &tran);
                    break;
                  }

                case POSM_PRN_CHASE_COMPLETE:
                  {
                    ProcessChaseComplete(&qmsg, &tran);
                    break;
                  }

                case POSM_PRN_PRINTER_ERROR:
                  {
                    ProcessPrinterError(&qmsg);
                    break;
                  }

                case POSM_TILL_OPENED:
                  {
                    ProcessTillOpened(&till_open_tm);
                    break;
                  }

                case POSM_TILL_CLOSED:
                  {
                    ProcessTillClosed(&tran);
                    break;
                  }

                case POSM_SYS_DEVICE_ONLINE:
                  {
                    DeviceStatus(&qmsg);
                    OpenDevice(&qmsg);

                    break;
                  }
                case POSM_SYS_DEVICE_OFFLINE:
                  {
                    DeviceStatus(&qmsg);
                    break;
                  }
                default:
                  {
                    DeviceStatus(&qmsg);
                    break;
                  }
              }
          }

        if (exit_switch == FALSE)
          {
            if ((pos_state == Signed_Off_State) && welcome_displayed == FALSE)
              {
                welcome_displayed = TRUE;

                /*
                 * Write the initial message to the display
                 */
                sleep(2);
                DisplayFullScreen(devInfo[DSP].devdes, achGreeting, 0);
              }
            else if (pos_state == Till_Opened_State)
              {
                if (time((time_t *) 0) >= till_open_tm)
                  {
                    /* Write to Display */
                    DisplayFullScreen(devInfo[DSP].devdes,
                                      "Close Cash Drawer", 0);

                    SoundTone();

                    till_open_tm += 60;
                  }
              }

            if (rc == -1)
              {
                if ((coerr = PosErrno()) == POSERR_SYS_INTERRUPTED)
                  {
                    /**********************************************/
                    /* Interrupt key pressed from main keyboard.  */
                    /* Go through the closing routines.           */
                    /**********************************************/
                    EndProgramKey();
                  }
                else if (coerr != POSERR_SYS_TIMEOUT)
                  {
                    fprintf(stderr, "PosRead queue: errno = %s\n",
                                    GetTableEntry(ErrnoTable, coerr));
                  }
              }
          }
      }

    fclose(item_rec_file);

#if defined(POSTOUCH)
    if (devInfo[TCH].devdes > 0)
      EndTouchScreen();
#endif

    ReleaseAll(devInfo, num_devices);

    printf("***** devices released *****\n");

    CloseAllDevices(devInfo, num_devices);
    printf("***** devices closed *****\n");

    printf("***** %s ending *****\n\n", argv[0]);

    return 0;
}

